#!/usr/bin/env python
"""Apply a cut on a table & copy to a new file.

You can use pytable's expression syntax to apply logical conditions
like

    sqrt(pos_x**2 + pos_y**2) <= 150.0

using numexpr-like syntax to filter events from a table.
To only grab the events which fullfill that condition, use `my_table.where()`,
or the related functions `read_where` and `append_where`.
"""

import sys
import tables as tb

infile = sys.argv[-2]
outfile = sys.argv[-1]

h5in = tb.open_file(infile, "r")
h5out = tb.open_file(outfile, "a")

tab_in = h5in.root.mva
tab_desc = tab_in.description
filt = tb.Filters(complevel=5)
tab_out = h5out.create_table("/", "mva", description=tab_desc, filters=filt)

expr = "sqrt(tf_r_x**2 + tf_r_y**2) <= 150.0"

tab_in.append_where(tab_out, expr)
tab_in.close()
tab_out.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""
================================
K40 Calibration Batch Processing
================================

Standalone job submitter for K40 offline calibrations with KM3Pipe.

.. code-block:: console

    Usage:
        qk40calib.py OUTPUT_PATH [options]
        qk40calib.py (-h | --help)

    Options:
        OUTPUT_PATH  Folder to store the calibration data.
        -d DET_ID    Detector ID [default: 29].
        -t TMAX      Coincidence time window [default: 10].
        -n N_RUNS    Number of runs to process per job [default: 10].
        -e ET        Estimated walltime per run in minutes [default: 8].
        -m VMEM      Estimated vmem for a job [default: 8G].
        -s RUNSETUP  Match [default: PHYS.1710v5-TUNED.HRV19.3D_T_S_MX.NBMODULE].
        -j JOBNAME   The name of the submitted jobs [default: k40calib].
        -l LOG_PATH  Path of the job log files [default: qlogs].
        -q           Dryrun: don't submit jobs, just print the first job script.
        -h --help    Show this screen.

"""

import os
import re
from glob import glob
import time
from km3pipe.shell import qsub, Script
import km3pipe as kp
from docopt import docopt


def main():
    args = docopt(__doc__)

    DET_ID = int(args["-d"])
    TMAX = int(args["-t"])
    ET_PER_RUN = int(args["-e"]) * 60  # [s]
    RUNS_PER_JOB = int(args["-n"])
    VMEM = args["-m"]
    CWD = os.getcwd()
    LOG_PATH = args["-l"]
    JOB_NAME = args["-j"]
    CALIB_PATH = os.path.join(CWD, args["OUTPUT_PATH"])
    RUN_SUBSTR = args["-s"]
    DRYRUN = args["-q"]

    if not os.path.exists(CALIB_PATH):
        os.makedirs(CALIB_PATH)

    db = kp.db.DBManager()
    run_table = db.run_table(det_id=DET_ID)
    phys_run_table = run_table[run_table.RUNSETUPNAME.str.contains(RUN_SUBSTR)]
    phys_runs = set(phys_run_table.RUN)
    processed_runs = set(
        int(re.search("_\\d{8}_(\\d{8})", s).group(1))
        for s in glob(os.path.join(CALIB_PATH, "*.k40_cal.p"))
    )
    remaining_runs = list(phys_runs - processed_runs)
    print("Remaining runs: {}".format(remaining_runs))
    s = Script()

    for job_id, runs_chunk in enumerate(kp.tools.chunks(remaining_runs, RUNS_PER_JOB)):
        n_runs = len(runs_chunk)
        print("Preparing batch script for a chunk of {} runs.".format(len(runs_chunk)))
        s.add("cd $TMPDIR; mkdir -p $USER; cd $USER")
        for run in runs_chunk:
            s.add("echo Processing {}:".format(run))
            irods_path = kp.tools.irods_path(DET_ID, run)
            root_filename = os.path.basename(irods_path)
            calib_filename = root_filename + ".k40_cal.p"
            s.add("iget -v {}".format(irods_path))
            s.add(
                "CTMIN=$(JPrint -f {}|grep '^ctMin'|awk '{{print $2}}')".format(
                    root_filename
                )
            )
            s.add(
                "k40calib {} {} -t {} -c $CTMIN -o {}".format(
                    root_filename, DET_ID, TMAX, calib_filename
                )
            )
            s.add("cp {} {}".format(calib_filename, CALIB_PATH))
            s.add("rm -f {}".format(root_filename))
            s.add("rm -f {}".format(calib_filename))
            s.add("echo Run {} processed.".format(run))
            s.add("echo " + 42 * "=")

        walltime = time.strftime("%H:%M:%S", time.gmtime(ET_PER_RUN * n_runs))
        qsub(
            s,
            "{}_{}".format(JOB_NAME, job_id),
            walltime=walltime,
            vmem=VMEM,
            log_path=LOG_PATH,
            irods=True,
            dryrun=DRYRUN,
        )

        if DRYRUN:
            break

        s.clear()


if __name__ == "__main__":
    main()



"""
==================
zt-plots.
==================

This example shows how to create a zt-plot of a given DU and event ID.
"""

# Author: Tamas Gal <tgal@km3net.de>
# License: BSD-3

import matplotlib.pyplot as plt
import km3pipe as kp
import km3pipe.style

km3pipe.style.use("km3pipe")

DU = 26
EVENT_ID = 23
filename = "data/km3net_jul13_90m_muatm50T655.km3_v5r1.JTE_r2356.root.0-499.h5"
cal = kp.calib.Calibration(filename="data/km3net_jul13_90m_r1494_corrected.detx")
pump = kp.io.hdf5.HDF5Pump(filename=filename)

raw_hits = pump[EVENT_ID]["Hits"]
hits = cal.apply(raw_hits).conv_to("pandas")

fig, ax = plt.subplots()

ax.scatter(hits.time, hits.pos_z, label="hits")

ax.set_title("zt-plot of event {0} on DU{1}".format(EVENT_ID, DU))
ax.set_xlabel("time [ns]")
ax.set_ylabel("z [m]")



#!/usr/bin/env python3
from os.path import join, dirname

import km3pipe as kp
from km3pipe.testing import TestCase, data_path
from km3pipe.io.offline import EventPump, OfflinePump


class TestEventPump(TestCase):
    def setUp(self):
        self.pump = EventPump(
            filename=data_path(
                "offline/mcv5.0.DAT004340.propa.sirene.jte.jchain.aanet.4340.root"
            )
        )

    def test_iteration(self):
        i = 0
        for blob in self.pump:
            i += 1

        assert 3 == i

    def test_getitem(self):
        blob = self.pump[0]
        assert "Header" in blob
        assert "Hits" in blob
        assert "McHits" in blob
        assert "McTracks" in blob

        assert 86 == len(blob["Hits"])

        with self.assertRaises(IndexError):
            self.pump[4]

    def test_hits(self):
        n_hits = [86, 111, 83]
        i = 0
        for blob in self.pump:
            assert n_hits[i] == len(blob["Hits"])
            i += 1

    def test_mc_hits(self):
        n_mc_hits = [147, 291, 160]
        i = 0
        for blob in self.pump:
            assert n_mc_hits[i] == len(blob["McHits"])
            i += 1

    def test_mc_tracks(self):
        n_mc_tracks = [4, 8, 36]
        i = 0
        for blob in self.pump:
            assert n_mc_tracks[i] == len(blob["McTracks"])
            i += 1

    def test_header(self):
        for blob in self.pump:
            pass

        self.assertListEqual([0, 1000000000.0, -1, -0.052], list(blob["Header"].cut_in))


class TestOfflinePump(TestCase):
    def setUp(self):
        self.pump = OfflinePump(
            filename=data_path(
                "offline/mcv5.0.DAT004340.propa.sirene.jte.jchain.aanet.4340.root"
            )
        )

    def test_offline_pump_iteration(self):
        i = 0
        for blob in self.pump:
            i += 1
        assert 3 == i

    def test_offline_pump_has_header(self):
        for blob in self.pump:
            assert "header" in blob

    def test_offline_pump_in_pipeline(self):
        class Observer(kp.Module):
            def configure(self):
                self.i = 0

            def process(self, blob):
                assert "header" in blob
                assert "event" in blob
                self.i += 1
                return blob

            def finish(self):
                return {"n_events": self.i}

        pipe = kp.Pipeline()
        pipe.attach(
            OfflinePump,
            filename=data_path(
                "offline/mcv5.0.DAT004340.propa.sirene.jte.jchain.aanet.4340.root"
            ),
        )
        pipe.attach(Observer)
        results = pipe.drain()

        assert 3 == results["Observer"]["n_events"]



__author__ = "tamasgal"

from km3pipe import Module, Pipeline
from km3pipe.io import EvtPump


class PrintBlob(Module):
    def process(self, blob):
        print(blob.keys())
        return blob


pipeline = Pipeline()
pipeline.attach(EvtPump, "evtpump", filename="files/example_numuNC.evt")
pipeline.attach(PrintBlob, "printer")
pipeline.drain(1)



# Filename: mc.py
# pylint: disable=C0103
"""
Monte Carlo related things.

"""
import particle
import numpy as np

from .logger import get_logger

__author__ = "Tamas Gal and Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = get_logger(__name__)  # pylint: disable=C0103


def geant2pdg(geant_code):
    """Convert GEANT particle ID to PDG"""
    try:
        return particle.Geant3ID(geant_code).to_pdgid()
    except KeyError:
        return 0


def pdg2name(pdg_id):
    """Convert PDG ID to human readable names"""
    try:
        return particle.Particle.from_pdgid(pdg_id).name
    except particle.particle.particle.InvalidParticle:
        log.debug("No particle with PDG ID %d found!" % pdg_id)
        return "N/A"


def name2pdg(name):
    """Return best match of a PDG ID for the given name"""
    return particle.Particle.from_string(name)


def most_energetic(df):
    """Grab most energetic particle from mc_tracks dataframe."""
    idx = df.groupby(["event_id"])["energy"].transform(max) == df["energy"]
    return df[idx].reindex()


def leading_particle(df):
    """Grab leading particle (neutrino, most energetic bundle muon).

    Note: selecting the most energetic mc particle does not always select
    the neutrino! In some sub-percent cases, the post-interaction
    secondaries can have more energy than the incoming neutrino!

    aanet convention: mc_tracks[0] = neutrino
    so grab the first row

    if the first row is not unique (neutrinos are unique), it's a muon bundle
    grab the most energetic then
    """
    leading = df.groupby("event_id", as_index=False).first()
    unique = leading.type.unique()

    if len(unique) == 1 and unique[0] == 0:
        leading = most_energetic(df)
    return leading


def get_flavor(pdg_types):
    """Build a 'flavor' from the 'type' column."""
    import pandas as pd

    return pd.Series(pdg_types).apply(pdg2name)


def _p_eq_nu(pdg_type):
    return np.abs(pdg_type) in {12, 14, 16}


def _p_eq_mu(pdg_type):
    return pdg_type == -13


def is_neutrino(pdg_types):
    """flavor string -> is_neutrino"""
    import pandas as pd

    return pd.Series(pdg_types).apply(_p_eq_nu)


def is_muon(pdg_types):
    """flavor string -> is_neutrino"""
    import pandas as pd

    return pd.Series(pdg_types).apply(_p_eq_mu)



# Filename: dataclasses.py
# pylint: disable=W0232,C0103,C0111
# vim:set ts=4 sts=4 sw=4 et syntax=python:
"""
Dataclasses for internal use. Heavily based on Numpy arrays.
"""
import numpy as np

__author__ = "Tamas Gal and Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Moritz Lotze"
__email__ = "mlotze@km3net.de"
__status__ = "Development"
__all__ = ("TEMPLATES",)

TEMPLATES = {
    "Direction": {
        "dtype": np.dtype([("dir_x", "<f4"), ("dir_y", "<f4"), ("dir_z", "<f4"),]),
        "h5loc": None,
        "split_h5": False,
        "h5singleton": False,
    },
    "Position": {
        "dtype": np.dtype([("pos_x", "<f4"), ("pos_y", "<f4"), ("pos_z", "<f4"),]),
        "h5loc": None,
        "split_h5": False,
        "h5singleton": False,
    },
    "EventInfo": {
        "dtype": np.dtype(
            [
                ("det_id", "<i4"),
                ("frame_index", "<u4"),
                ("livetime_sec", "<u8"),
                ("mc_id", "<i4"),
                ("mc_t", "<f8"),
                ("n_events_gen", "<u8"),
                ("n_files_gen", "<u8"),
                ("overlays", "<u4"),
                ("trigger_counter", "<u8"),
                ("trigger_mask", "<u8"),
                ("utc_nanoseconds", "<u8"),
                ("utc_seconds", "<u8"),
                ("weight_w1", "<f8"),
                ("weight_w2", "<f8"),
                ("weight_w3", "<f8"),
                ("run_id", "<u8"),
                ("group_id", "<u8"),
            ]
        ),
        "h5loc": "/event_info",
        "split_h5": False,
        "h5singleton": False,
    },
    "TimesliceHits": {
        "dtype": np.dtype(
            [
                ("channel_id", "u1"),
                ("dom_id", "<u4"),
                ("time", "<f8"),
                ("tot", "u1"),
                ("group_id", "<u4"),
            ]
        ),
        "h5loc": "/timeslice_hits",
        "split_h5": True,
        "h5singleton": False,
    },
    "Hits": {
        "dtype": np.dtype(
            [
                ("channel_id", "u1"),
                ("dom_id", "<u4"),
                ("time", "<f8"),
                ("tot", "u1"),
                ("triggered", "?"),
                ("group_id", "<u4"),
            ]
        ),
        "h5loc": "/hits",
        "split_h5": True,
        "h5singleton": False,
    },
    "CalibHits": {
        "dtype": np.dtype(
            [
                ("channel_id", "u1"),
                ("dir_x", "<f4"),
                ("dir_y", "<f4"),
                ("dir_z", "<f4"),
                ("dom_id", "<u4"),
                ("du", "u1"),
                ("floor", "u1"),
                ("pos_x", "<f4"),
                ("pos_y", "<f4"),
                ("pos_z", "<f4"),
                ("t0", "<f4"),
                ("time", "<f8"),
                ("tot", "u1"),
                ("triggered", "?"),
                ("group_id", "<u4"),
            ]
        ),
        "h5loc": "/hits",
        "split_h5": True,
        "h5singleton": False,
    },
    "McHits": {
        "dtype": np.dtype(
            [
                ("a", "<f4"),
                ("origin", "<u4"),
                ("pmt_id", "<u4"),
                ("time", "<f8"),
                ("group_id", "<u4"),
            ]
        ),
        "h5loc": "/mc_hits",
        "split_h5": True,
        "h5singleton": False,
    },
    "CalibMcHits": {
        "dtype": np.dtype(
            [
                ("a", "<f4"),
                ("dir_x", "<f4"),
                ("dir_y", "<f4"),
                ("dir_z", "<f4"),
                ("origin", "<u4"),
                ("pmt_id", "<u4"),
                ("pos_x", "<f4"),
                ("pos_y", "<f4"),
                ("pos_z", "<f4"),
                ("time", "<f8"),
                ("group_id", "<u4"),
            ]
        ),
        "h5loc": "/mc_hits",
        "split_h5": True,
        "h5singleton": False,
    },
    "Tracks": {
        "dtype": np.dtype(
            [
                ("bjorkeny", "<f8"),
                ("dir_x", "<f8"),
                ("dir_y", "<f8"),
                ("dir_z", "<f8"),
                ("energy", "<f8"),
                ("id", "<u4"),
                ("interaction_channel", "<u4"),
                ("is_cc", "<u4"),  # TODO: consider bool ('?') for slicing
                ("length", "<f8"),
                ("pos_x", "<f8"),
                ("pos_y", "<f8"),
                ("pos_z", "<f8"),
                ("time", "<i4"),
                ("type", "<i4"),
                ("group_id", "<u4"),
            ]
        ),
        "h5loc": "/tracks",
        "split_h5": False,
        "h5singleton": False,
    },
    "McTracks": {
        "dtype": np.dtype(
            [
                ("bjorkeny", "<f8"),
                ("dir_x", "<f8"),
                ("dir_y", "<f8"),
                ("dir_z", "<f8"),
                ("energy", "<f8"),
                ("id", "<u4"),
                ("interaction_channel", "<u4"),
                ("is_cc", "<u4"),  # TODO: consider bool ('?') for slicing
                ("length", "<f8"),
                ("pos_x", "<f8"),
                ("pos_y", "<f8"),
                ("pos_z", "<f8"),
                ("time", "<i4"),
                ("type", "<i4"),
                ("group_id", "<u4"),
            ]
        ),
        "h5loc": "/mc_tracks",
        "split_h5": False,
        "h5singleton": False,
    },
    "SummaryFrameInfo": {
        "dtype": np.dtype(
            [
                ("dom_id", "<u4"),
                ("fifo_status", "<u4"),
                ("frame_id", "<u4"),
                ("frame_index", "<u4"),
                ("has_udp_trailer", "<u4"),
                ("high_rate_veto", "<u4"),
                ("max_sequence_number", "<u4"),
                ("n_packets", "<u4"),
                ("slice_id", "<u4"),
                ("utc_nanoseconds", "<u4"),
                ("utc_seconds", "<u4"),
                ("white_rabbit_status", "<u4"),
            ]
        ),
        "h5loc": "/summary_slice_info",
        "split_h5": False,
        "h5singleton": False,
    },
    "SummarysliceInfo": {
        "dtype": np.dtype(
            [
                ("frame_index", "<u4"),
                ("slice_id", "<u4"),
                ("timestamp", "<u4"),
                ("nanoseconds", "<u4"),
                ("n_frames", "<u4"),
            ]
        ),
        "h5loc": "/todo",
        "split_h5": False,
        "h5singleton": False,
    },
    "TimesliceInfo": {
        "dtype": np.dtype(
            [
                ("frame_index", "<u4"),
                ("slice_id", "<u4"),
                ("timestamp", "<u4"),
                ("nanoseconds", "<u4"),
                ("n_frames", "<u4"),
            ]
        ),
        "h5loc": "/timeslice_info",
        "split_h5": False,
        "h5singleton": False,
    },
    "SummaryframeSeries": {
        "dtype": np.dtype(
            [
                ("dom_id", "<u4"),
                ("max_sequence_number", "<u4"),
                ("n_received_packets", "<u4"),
                ("group_id", "<u4"),
            ]
        ),
        "h5loc": "/todo",
        "split_h5": False,
        "h5singleton": False,
    },
    "TimesliceFrameInfo": {
        "dtype": np.dtype(
            [
                ("det_id", "<i4"),
                ("run_id", "<u8"),
                ("sqnr", "<u8"),
                ("timestamp", "<u4"),
                ("nanoseconds", "<u4"),
                ("dom_id", "<u4"),
                ("dataqueue_status", "<u4"),
                ("dom_status", "<u4"),
                ("n_hits", "<u4"),
            ]
        ),
        "h5loc": "/todo",
        "split_h5": False,
        "h5singleton": False,
    },
    "SummaryFrameSeries": {
        "dtype": np.dtype(
            [
                ("dom_id", "<u4"),
                ("max_sequence_number", "<u4"),
                ("n_received_packets", "<u4"),
                ("group_id", "<u4"),
            ]
        ),
        "h5loc": "/summary_frame_series",
        "split_h5": False,
        "h5singleton": False,
    },
}



#!/usr/bin/env python
# vim: ts=4 sw=4 et

from datetime import datetime
from collections import deque
import os
from io import StringIO
import shutil
import time
import threading

import matplotlib
# Force matplotlib to not use any Xwindows backend.
matplotlib.use('Agg')    # noqa
import matplotlib.pyplot as plt
import matplotlib.dates as md
from matplotlib.colors import LogNorm
import numpy as np

from km3pipe import Pipeline, Module
from km3pipe.calib import Calibration
from km3pipe.io import CHPump
from km3pipe.io.daq import (DAQProcessor, DAQPreamble, DAQEvent)
import km3pipe.style

km3pipe.style.use('km3pipe')

PLOTS_PATH = 'www/plots'
N_DOMS = 18
N_DUS = 2
cal = Calibration(det_id=14)
detector = cal.detector

xfmt = md.DateFormatter('%Y-%m-%d %H:%M')
lock = threading.Lock()


class DOMHits(Module):
    def configure(self):
        self.run = True
        self.max_events = 1000
        self.hits = deque(maxlen=1000)
        self.triggered_hits = deque(maxlen=1000)
        self.thread = threading.Thread(target=self.plot).start()

    def process(self, blob):
        tag = str(blob['CHPrefix'].tag)

        if not tag == 'IO_EVT':
            return blob

        data = blob['CHData']
        data_io = StringIO(data)
        preamble = DAQPreamble(file_obj=data_io)    # noqa
        event = DAQEvent(file_obj=data_io)
        with lock:
            hits = np.zeros(N_DOMS * N_DUS)
            for dom_id, _, _, _ in event.snapshot_hits:
                du, floor, _ = detector.doms[dom_id]
                hits[(du - 1) * N_DOMS + floor - 1] += 1
            self.hits.append(hits)
            triggered_hits = np.zeros(N_DOMS * N_DUS)
            for dom_id, _, _, _, _ in event.triggered_hits:
                du, floor, _ = detector.doms[dom_id]
                triggered_hits[(du - 1) * N_DOMS + floor - 1] += 1
            self.triggered_hits.append(triggered_hits)

        return blob

    def plot(self):
        while self.run:
            with lock:
                self.create_plots()
            time.sleep(50)

    def create_plots(self):
        if len(self.hits) > 0:
            self.create_plot(self.hits, "Hits on DOMs", 'hits_on_doms')
        if len(self.triggered_hits) > 0:
            self.create_plot(
                self.triggered_hits, "Triggered Hits on DOMs",
                'triggered_hits_on_doms'
            )

    def create_plot(self, hits, title, filename):
        fig, ax = plt.subplots(figsize=(16, 8))
        ax.grid(True)
        ax.set_axisbelow(True)
        hit_matrix = np.array([np.array(x) for x in hits]).transpose()
        im = ax.matshow(
            hit_matrix,
            interpolation='nearest',
            filternorm=None,
            cmap='plasma',
            aspect='auto',
            origin='lower',
            zorder=3,
            norm=LogNorm(vmin=1, vmax=np.amax(hit_matrix))
        )
        yticks = np.arange(N_DOMS * N_DUS)
        ytick_labels = [
            "DU{0:0.0f}-DOM{1:02d}".format(
                np.ceil((y + 1) / N_DOMS), y % (N_DOMS) + 1
            ) for y in yticks
        ]
        ax.set_yticks(yticks)
        ax.set_yticklabels(ytick_labels)
        ax.tick_params(labelbottom=False)
        ax.tick_params(labeltop=False)
        ax.set_xlabel("event (latest on the right)")
        ax.set_title(
            "{0} - via the last {1} Events\n{2} UTC".format(
                title, self.max_events,
                datetime.utcnow().strftime("%c")
            )
        )
        cb = fig.colorbar(im, pad=0.05)
        cb.set_label("number of hits")

        fig.tight_layout()

        f = os.path.join(PLOTS_PATH, filename + '.png')
        f_tmp = os.path.join(PLOTS_PATH, filename + '_tmp.png')
        plt.savefig(f_tmp, dpi=120, bbox_inches="tight")
        plt.close('all')
        shutil.move(f_tmp, f)

    def finish(self):
        self.run = False
        if self.thread is not None:
            self.thread.stop()


pipe = Pipeline()
pipe.attach(
    CHPump,
    host='127.0.0.1',
    port=5553,
    tags='IO_EVT, IO_SUM',
    timeout=60 * 60 * 24 * 7,
    max_queue=2000
)
pipe.attach(DAQProcessor)
pipe.attach(DOMHits)
pipe.drain()



# -*- coding: utf-8 -*-
"""
=============================
Reading and Parsing EVT files
=============================

This example shows how to read and parse EVT files, which are used in our
Monte Carlo productions.
"""

# Author: Tamas Gal <tgal@km3net.de>, Moritz Lotze >mlotze@km3net.de>
# License: BSD-3
import matplotlib.pyplot as plt
import numpy as np

from km3modules.common import StatusBar
from km3pipe import Module, Pipeline
from km3pipe.dataclasses import Table
from km3pipe.calib import Calibration
from km3pipe.io import EvtPump
from km3pipe.math import pld3
import km3pipe.style

km3pipe.style.use("km3pipe")

filename = "../data/numu_cc.evt"
detx = "../data/km3net_jul13_90m_r1494_corrected.detx"


class VertexHitDistanceCalculator(Module):
    """Calculate vertex-hit-distances"""

    def configure(self):
        self.distances = []

    def process(self, blob):
        tracks = blob["TrackIns"]
        muons = tracks[tracks.type == 5]
        muon = Table(muons[np.argmax(muons.energy)])
        hits = blob["CalibHits"]
        dist = pld3(hits.pos, muon.pos, muon.dir)
        self.distances.append(dist)
        return blob

    def finish(self):
        dist_flat = np.concatenate(self.distances)
        plt.hist(dist_flat)
        plt.savefig("dists.png")


pipe = Pipeline()
pipe.attach(EvtPump, filename=filename, parsers=["km3"])
pipe.attach(StatusBar, every=100)
pipe.attach(Calibration, filename=detx)
pipe.attach(VertexHitDistanceCalculator)
pipe.drain(5)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""
==================
Monitoring Channel
==================

This application demonstrates how to access monitoring channel data.

"""

# Author: Tamas Gal <tgal@km3net.de>
# License: MIT
#!/usr/bin/env python
# vim: ts=4 sw=4 et
import io
import km3pipe as kp

__author__ = "Tamas Gal"
__email__ = "tgal@km3net.de"


def print_monitoring_data(blob):
    data = io.BytesIO(blob["CHData"])
    print(kp.io.daq.TMCHData(data))
    return blob


pipe = kp.Pipeline()
pipe.attach(kp.io.CHPump, host="localhost", tags="IO_MONIT")
pipe.attach(print_monitoring_data)
pipe.drain()



# Filename: no_test_daq.py
# pylint: disable=C0111,R0904,C0103
"""
Tests for KM3NeT binary formats readout.

"""
from os.path import dirname, join

import numpy as np

from km3pipe.testing import TestCase, data_path
from km3pipe.io.daq import (
    DAQPump,
    DAQPreamble,
    DAQHeader,
    DAQSummaryslice,
    DMMonitor,
    TMCHRepump,
    TimesliceParser,
)


class TestDAQPump(TestCase):
    def test_init_with_filename(self):
        DAQPump(filename=data_path("daq/IO_SUM.dat"))

    def test_frame_positions_in_io_sum(self):
        p = DAQPump(filename=data_path("daq/IO_SUM.dat"))
        assert 81 == len(p.frame_positions)
        self.assertListEqual([0, 656, 1312], p.frame_positions[:3])
        self.assertListEqual([50973, 51629, 52285], p.frame_positions[-3:])

    def test_frame_positions_in_io_evt(self):
        p = DAQPump(filename=data_path("daq/IO_EVT.dat"))
        assert 38 == len(p.frame_positions)
        self.assertListEqual([0, 570, 986], p.frame_positions[:3])
        self.assertListEqual([13694, 14016, 14360], p.frame_positions[-3:])

    def test_blob_in_io_sum(self):
        p = DAQPump(filename=data_path("daq/IO_SUM.dat"))
        blob = p.next_blob()
        assert "DAQSummaryslice" in blob.keys()
        assert "DAQPreamble" in blob.keys()
        assert "DAQHeader" in blob.keys()
        assert 16 == blob["DAQSummaryslice"].n_summary_frames

    def test_blob_in_io_evt(self):
        p = DAQPump(filename=data_path("daq/IO_EVT.dat"))
        blob = p.next_blob()
        assert "DAQEvent" in blob.keys()
        assert "DAQPreamble" in blob.keys()
        assert "DAQHeader" in blob.keys()
        event = blob["DAQEvent"]
        assert 13 == event.n_triggered_hits
        assert 28 == event.n_snapshot_hits

    def test_blob_iteration(self):
        p = DAQPump(filename=data_path("daq/IO_EVT.dat"))
        for blob in p:
            pass

    def test_get_item(self):
        p = DAQPump(filename=data_path("daq/IO_EVT.dat"))
        blob = p[4]
        event = blob["DAQEvent"]
        assert 6 == event.n_triggered_hits
        assert 17 == event.n_snapshot_hits


class TestTMCHRepump(TestCase):
    def test_reading_version_2(self):
        repump = TMCHRepump(filename=data_path("daq/IO_MONIT.dat"))
        packets = [p["TMCHData"] for p in repump]

        p1 = packets[0]
        p2 = packets[5]

        assert 86 == p1.run
        assert 0 == p1.udp_sequence_number
        assert 541 == p1.utc_seconds
        assert 500000000 == p1.nanoseconds
        assert 806472270 == p1.dom_id
        assert 2 == p1.version
        self.assertAlmostEqual(199.05982971191406, p1.yaw)
        self.assertAlmostEqual(0.5397617816925049, p1.pitch)
        self.assertAlmostEqual(-0.2243121862411499, p1.roll)
        self.assertAlmostEqual(32.35, p1.temp)
        self.assertAlmostEqual(16.77, p1.humidity)
        assert np.allclose(np.full(31, 0), p1.pmt_rates)
        assert np.allclose([0.00708725, 0.00213623, -0.86456668], p1.A)
        assert np.allclose([-0.2621212, 0.02363636, 0.1430303], p1.H)
        assert np.allclose([-2.87721825, -0.83284622, -0.28969574], p1.G)

        assert 86 == p2.run
        assert 0 == p2.udp_sequence_number
        assert 542 == p2.utc_seconds
        assert 0 == p2.nanoseconds
        assert 806472270 == p2.dom_id
        assert 2 == p2.version
        self.assertAlmostEqual(199.05982971191406, p2.yaw)
        self.assertAlmostEqual(0.5397617816925049, p2.pitch)
        self.assertAlmostEqual(-0.2243121862411499, p2.roll)
        self.assertAlmostEqual(32.35, p2.temp)
        self.assertAlmostEqual(16.77, p2.humidity)
        assert np.allclose(np.full(31, 0), p2.pmt_rates)
        assert np.allclose([0.00708725, 0.00213623, -0.86456668], p2.A)
        assert np.allclose([-0.2621212, 0.02363636, 0.1430303], p2.H)
        assert np.allclose([-2.87721825, -0.83284622, -0.28969574], p2.G)


class TestDMMonitor(TestCase):
    def test_init(self):
        dmm = DMMonitor("a")
        assert "http://a:1302/mon/" == dmm._url

    def test_available_parameters(self):
        dmm = DMMonitor("a")
        dmm._available_parameters = ["b", "c"]
        self.assertListEqual(["b", "c"], dmm.available_parameters)


class TestTimesliceParser(TestCase):
    def test_l0(self):
        with open(data_path("daq/IO_TSL0.dat"), "rb") as fobj:
            ts_info, ts_frameinfos, ts_hits = TimesliceParser()._parse_timeslice(fobj)
        assert 200 == len(ts_hits)
        assert 25 == ts_hits[0].tot
        assert 808447031 == ts_hits[23].dom_id
        assert 232 == ts_info[0].frame_index

    def test_l1(self):
        with open(data_path("daq/IO_TSL1.dat"), "rb") as fobj:
            ts_info, ts_frameinfos, ts_hits = TimesliceParser()._parse_timeslice(fobj)
        assert 0 == len(ts_hits)
        assert 4873 == ts_info[0].frame_index

    def test_l2(self):
        with open(data_path("daq/IO_TSL2.dat"), "rb") as fobj:
            ts_info, ts_frameinfos, ts_hits = TimesliceParser()._parse_timeslice(fobj)
        assert 0 == len(ts_hits)
        assert 4872 == ts_info[0].frame_index



# Filename: dataclasses.py
# pylint: disable=W0232,C0103,C0111
# vim:set ts=4 sts=4 sw=4 et syntax=python:
"""
Dataclasses for internal use. Heavily based on Numpy arrays.
"""
import itertools

import numpy as np
from numpy.lib import recfunctions as rfn

from .dataclass_templates import TEMPLATES
from .logger import get_logger
from .tools import istype

__author__ = "Tamas Gal and Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"
__all__ = ("Table", "is_structured", "has_structured_dt", "inflate_dtype")

DEFAULT_H5LOC = "/misc"
DEFAULT_NAME = "Generic Table"
DEFAULT_SPLIT = False
DEFAULT_H5SINGLETON = False

log = get_logger(__name__)


def has_structured_dt(arr):
    """Check if the array representation has a structured dtype."""
    arr = np.asanyarray(arr)
    return is_structured(arr.dtype)


def is_structured(dt):
    """Check if the dtype is structured."""
    if not hasattr(dt, "fields"):
        return False
    return dt.fields is not None


def inflate_dtype(arr, names):
    """Create structured dtype from a 2d ndarray with unstructured dtype."""
    arr = np.asanyarray(arr)
    if has_structured_dt(arr):
        return arr.dtype
    s_dt = arr.dtype
    dt = [(n, s_dt) for n in names]
    dt = np.dtype(dt)
    return dt


class Table(np.recarray):
    """2D generic Table with grouping index.

    This is a `np.recarray` subclass with some metadata and helper methods.

    You can initialize it directly from a structured numpy array,
    a pandas DataFrame, a dictionary of (columnar) arrays; or, initialize it
    from a list of rows/list of columns using the appropriate factory.

    This class adds the following to ``np.recarray``:

    Parameters
    ----------
    data: array-like or dict(array-like)
        numpy array with structured/flat dtype, or dict of arrays.
    h5loc: str
        Location in HDF5 file where to store the data. [default: '/misc']
    h5singleton: bool
        Tables defined as h5singletons are only written once to an HDF5 file.
        This is used for headers for example (default=False).
    dtype: numpy dtype
        Datatype over array. If not specified and data is an unstructured
        array, ``names`` needs to be specified. [default: None]

    Attributes
    ----------
    h5loc: str
        HDF5 group where to write into. (default='/misc')
    split_h5: bool
        Split the array into separate arrays, column-wise, when saving
        to hdf5? (default=False)
    name: str
        Human-readable name, e.g. 'Hits'
    h5singleton: bool
        Tables defined as h5singletons are only written once to an HDF5 file.
        This is used for headers for example (default=False).

    Methods
    -------
    from_dict(arr_dict, dtype=None, **kwargs)
        Create an Table from a dict of arrays (similar to pandas).
    from_template(data, template, **kwargs)
        Create an array from a dict of arrays with a predefined dtype.
    sorted(by)
        Sort the table by one of its columns.
    append_columns(colnames, values)
        Append new columns to the table.
    to_dataframe()
        Return as pandas dataframe.
    from_dataframe(df, **kwargs)
        Instantiate from a dataframe.
    from_rows(list_of_rows, **kwargs)
        Instantiate from an array-like with shape (n_rows, n_columns).
    from_columns(list_of_columns, **kwargs)
        Instantiate from an array-like with shape (n_columns, n_rows).
    """

    def __new__(
        cls,
        data,
        h5loc=DEFAULT_H5LOC,
        dtype=None,
        split_h5=DEFAULT_SPLIT,
        name=DEFAULT_NAME,
        h5singleton=DEFAULT_H5SINGLETON,
        **kwargs
    ):
        if isinstance(data, dict):
            return cls.from_dict(
                data,
                h5loc=h5loc,
                dtype=dtype,
                split_h5=split_h5,
                name=name,
                h5singleton=h5singleton,
                **kwargs
            )
        if istype(data, "DataFrame"):
            return cls.from_dataframe(
                data,
                h5loc=h5loc,
                dtype=dtype,
                split_h5=split_h5,
                name=name,
                h5singleton=h5singleton,
                **kwargs
            )
        if isinstance(data, (list, tuple)):
            raise ValueError(
                "Lists/tuples are not supported! "
                "Please use the `from_rows` or `from_columns` method instead!"
            )
        if isinstance(data, np.record):
            # single record from recarrary/kp.Tables, let's blow it up
            data = data[np.newaxis]
        if not has_structured_dt(data):
            # flat (nonstructured) dtypes fail miserably!
            # default to `|V8` whyever
            raise ValueError(
                "Arrays without structured dtype are not supported! "
                "Please use the `from_rows` or `from_columns` method instead!"
            )

        if dtype is None:
            dtype = data.dtype

        assert is_structured(dtype)

        if dtype != data.dtype:
            dtype_names = set(dtype.names)
            data_dtype_names = set(data.dtype.names)
            if dtype_names == data_dtype_names:
                if not all(dtype[f] == data.dtype[f] for f in dtype_names):
                    log.critical(
                        "dtype mismatch! Matching field names but differing "
                        "field types, no chance to reorder.\n"
                        "dtype of data:   %s\n"
                        "requested dtype: %s" % (data.dtype, dtype)
                    )
                    raise ValueError("dtype mismatch")
                log.once(
                    "dtype mismatch, but matching field names and types. "
                    "Rordering input data...",
                    identifier=h5loc,
                )
                data = Table({f: data[f] for f in dtype_names}, dtype=dtype)
            else:
                log.critical(
                    "dtype mismatch, no chance to reorder due to differing "
                    "fields!\n"
                    "dtype of data:   %s\n"
                    "requested dtype: %s" % (data.dtype, dtype)
                )
                raise ValueError("dtype mismatch")

        obj = np.asanyarray(data, dtype=dtype).view(cls)
        obj.h5loc = h5loc
        obj.split_h5 = split_h5
        obj.name = name
        obj.h5singleton = h5singleton
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            # called from explicit contructor
            return obj
        # views or slices
        self.h5loc = getattr(obj, "h5loc", DEFAULT_H5LOC)
        self.split_h5 = getattr(obj, "split_h5", DEFAULT_SPLIT)
        self.name = getattr(obj, "name", DEFAULT_NAME)
        self.h5singleton = getattr(obj, "h5singleton", DEFAULT_H5SINGLETON)
        # attribute access returns void instances on slicing/iteration
        # kudos to
        # https://github.com/numpy/numpy/issues/3581#issuecomment-108957200
        if obj is not None and type(obj) is not type(self):
            self.dtype = np.dtype((np.record, obj.dtype))

    def __array_wrap__(self, out_arr, context=None):
        # then just call the parent
        return Table(
            np.recarray.__array_wrap__(self, out_arr, context),
            h5loc=self.h5loc,
            split_h5=self.split_h5,
            name=self.name,
            h5singleton=self.h5singleton,
        )

    @staticmethod
    def _expand_scalars(arr_dict):
        scalars = []
        maxlen = 1  # have at least 1-elem arrays
        for k, v in arr_dict.items():
            if np.isscalar(v):
                scalars.append(k)
                continue
            # TODO: this is not covered yet, don't know if we need this
            # if hasattr(v, 'shape') and v.shape == (1,):  # np.array([1])
            #     import pdb; pdb.set_trace()
            #     arr_dict[k] = v[0]
            #     continue
            if hasattr(v, "ndim") and v.ndim == 0:  # np.array(1)
                arr_dict[k] = v.item()
                continue
            if len(v) > maxlen:
                maxlen = len(v)
        for s in scalars:
            arr_dict[s] = np.full(maxlen, arr_dict[s])
        return arr_dict

    @classmethod
    def from_dict(cls, arr_dict, dtype=None, fillna=False, **kwargs):
        """Generate a table from a dictionary of arrays.
        """
        # i hope order of keys == order or values
        if dtype is None:
            names = sorted(list(arr_dict.keys()))
        else:
            dtype = np.dtype(dtype)
            dt_names = [f for f in dtype.names]
            dict_names = [k for k in arr_dict.keys()]
            missing_names = set(dt_names) - set(dict_names)
            if missing_names:
                if fillna:
                    dict_names = dt_names
                    for missing_name in missing_names:
                        arr_dict[missing_name] = np.nan
                else:
                    raise KeyError("Dictionary keys and dtype fields do not match!")
            names = list(dtype.names)

        arr_dict = cls._expand_scalars(arr_dict)
        data = [arr_dict[key] for key in names]
        return cls(np.rec.fromarrays(data, names=names, dtype=dtype), **kwargs)

    @classmethod
    def from_columns(cls, column_list, dtype=None, colnames=None, **kwargs):
        if dtype is None or not is_structured(dtype):
            # infer structured dtype from array data + column names
            if colnames is None:
                raise ValueError(
                    "Need to either specify column names or a "
                    "structured dtype when passing unstructured arrays!"
                )
            dtype = inflate_dtype(column_list, colnames)
            colnames = dtype.names
        if len(column_list) != len(dtype.names):
            raise ValueError("Number of columns mismatch between data and dtype!")
        data = {k: column_list[i] for i, k in enumerate(dtype.names)}
        return cls(data, dtype=dtype, colnames=colnames, **kwargs)

    @classmethod
    def from_rows(cls, row_list, dtype=None, colnames=None, **kwargs):
        if dtype is None or not is_structured(dtype):
            # infer structured dtype from array data + column names
            if colnames is None:
                raise ValueError(
                    "Need to either specify column names or a "
                    "structured dtype when passing unstructured arrays!"
                )
            dtype = inflate_dtype(row_list, colnames)
        # this *should* have been checked above, but do this
        # just to be sure in case I screwed up the logic above;
        # users will never see this, this should only show in tests
        assert is_structured(dtype)
        data = np.asanyarray(row_list).view(dtype)
        # drop useless 2nd dim
        data = data.reshape((data.shape[0],))
        return cls(data, **kwargs)

    @property
    def templates_avail(self):
        return sorted(list(TEMPLATES.keys()))

    @classmethod
    def from_template(cls, data, template):
        """Create a table from a predefined datatype.

        See the ``templates_avail`` property for available names.

        Parameters
        ----------
        data
            Data in a format that the ``__init__`` understands.
        template: str or dict
            Name of the dtype template to use from ``kp.dataclasses_templates``
            or a ``dict`` containing the required attributes (see the other
            templates for reference).
        """
        name = DEFAULT_NAME
        if isinstance(template, str):
            name = template
            table_info = TEMPLATES[name]
        else:
            table_info = template
        if "name" in table_info:
            name = table_info["name"]
        dt = table_info["dtype"]
        loc = table_info["h5loc"]
        split = table_info["split_h5"]
        h5singleton = table_info["h5singleton"]

        return cls(
            data,
            h5loc=loc,
            dtype=dt,
            split_h5=split,
            name=name,
            h5singleton=h5singleton,
        )

    @staticmethod
    def _check_column_length(values, n):
        values = np.atleast_2d(values)
        for v in values:
            if len(v) == n:
                continue
            else:
                raise ValueError(
                    "Trying to append more than one column, but "
                    "some arrays mismatch in length!"
                )

    def append_columns(self, colnames, values, **kwargs):
        """Append new columns to the table.

        When appending a single column, ``values`` can be a scalar or an
        array of either length 1 or the same length as this array (the one
        it's appended to). In case of multiple columns, values must have
        the shape ``list(arrays)``, and the dimension of each array
        has to match the length of this array.

        See the docs for ``numpy.lib.recfunctions.append_fields`` for an
        explanation of the remaining options.
        """
        n = len(self)
        if np.isscalar(values):
            values = np.full(n, values)

        values = np.atleast_1d(values)
        if not isinstance(colnames, str) and len(colnames) > 1:
            values = np.atleast_2d(values)
            self._check_column_length(values, n)

        if values.ndim == 1:
            if len(values) > n:
                raise ValueError("New Column is longer than existing table!")
            elif len(values) > 1 and len(values) < n:
                raise ValueError(
                    "New Column is shorter than existing table, "
                    "but not just one element!"
                )
            elif len(values) == 1:
                values = np.full(n, values[0])
        new_arr = rfn.append_fields(
            self, colnames, values, usemask=False, asrecarray=True, **kwargs
        )
        return self.__class__(
            new_arr,
            h5loc=self.h5loc,
            split_h5=self.split_h5,
            name=self.name,
            h5singleton=self.h5singleton,
        )

    def drop_columns(self, colnames, **kwargs):
        """Drop  columns from the table.

        See the docs for ``numpy.lib.recfunctions.drop_fields`` for an
        explanation of the remaining options.
        """
        new_arr = rfn.drop_fields(
            self, colnames, usemask=False, asrecarray=True, **kwargs
        )
        return self.__class__(
            new_arr,
            h5loc=self.h5loc,
            split_h5=self.split_h5,
            name=self.name,
            h5singleton=self.h5singleton,
        )

    def sorted(self, by, **kwargs):
        """Sort array by a column.

        Parameters
        ==========
        by: str
            Name of the columns to sort by(e.g. 'time').
        """
        sort_idc = np.argsort(self[by], **kwargs)
        return self.__class__(
            self[sort_idc], h5loc=self.h5loc, split_h5=self.split_h5, name=self.name
        )

    def to_dataframe(self):
        from pandas import DataFrame

        return DataFrame(self)

    @classmethod
    def from_dataframe(cls, df, **kwargs):
        rec = df.to_records(index=False)
        return cls(rec, **kwargs)

    @classmethod
    def merge(cls, tables, fillna=False):
        """Merge a list of tables"""
        cols = set(itertools.chain(*[table.dtype.descr for table in tables]))

        tables_to_merge = []
        for table in tables:
            missing_cols = cols - set(table.dtype.descr)

            if missing_cols:
                if fillna:
                    n = len(table)
                    n_cols = len(missing_cols)
                    col_names = []
                    for col_name, col_dtype in missing_cols:
                        if "f" not in col_dtype:
                            raise ValueError(
                                "Cannot create NaNs for non-float"
                                " type column '{}'".format(col_name)
                            )
                        col_names.append(col_name)

                    table = table.append_columns(
                        col_names, np.full((n_cols, n), np.nan)
                    )
                else:
                    raise ValueError(
                        "Table columns do not match. Use fill_na=True"
                        " if you want to append missing values with NaNs"
                    )
            tables_to_merge.append(table)

        first_table = tables_to_merge[0]

        merged_table = sum(tables_to_merge[1:], first_table)

        merged_table.h5loc = first_table.h5loc
        merged_table.h5singleton = first_table.h5singleton
        merged_table.split_h5 = first_table.split_h5
        merged_table.name = first_table.name

        return merged_table

    def __add__(self, other):
        cols1 = set(self.dtype.descr)
        cols2 = set(other.dtype.descr)
        if len(cols1 ^ cols2) != 0:
            cols1 = set(self.dtype.names)
            cols2 = set(other.dtype.names)
            if len(cols1 ^ cols2) == 0:
                raise NotImplementedError
            else:
                raise TypeError("Table columns do not match")
        col_order = list(self.dtype.names)
        ret = self.copy()
        len_self = len(self)
        len_other = len(other)
        final_length = len_self + len_other
        ret.resize(final_length, refcheck=False)
        ret[len_self:] = other[col_order]
        return Table(
            ret,
            h5loc=self.h5loc,
            h5singleton=self.h5singleton,
            split_h5=self.split_h5,
            name=self.name,
        )

    def __str__(self):
        name = self.name
        spl = "split" if self.split_h5 else "no split"
        s = "{} {}\n".format(name, type(self))
        s += "HDF5 location: {} ({})\n".format(self.h5loc, spl)
        s += "\n".join(
            map(
                lambda d: "{2} (dtype: {1}) = {0}".format(self[d[0]], *d),
                self.dtype.descr,
            )
        )
        return s

    def __repr__(self):
        s = "{} {} (rows: {})".format(self.name, type(self), self.size)
        return s

    def __contains__(self, elem):
        return elem in self.dtype.names

    @property
    def pos(self):
        return np.array([self.pos_x, self.pos_y, self.pos_z]).T

    @pos.setter
    def pos(self, arr):
        arr = np.atleast_2d(arr)
        assert arr.shape[1] == 3
        assert len(arr) == len(self)
        self.pos_x = arr[:, 0]
        self.pos_y = arr[:, 1]
        self.pos_z = arr[:, 2]

    @property
    def dir(self):
        return np.array([self.dir_x, self.dir_y, self.dir_z]).T

    @dir.setter
    def dir(self, arr):
        arr = np.atleast_2d(arr)
        assert arr.shape[1] == 3
        assert len(arr) == len(self)
        self.dir_x = arr[:, 0]
        self.dir_y = arr[:, 1]
        self.dir_z = arr[:, 2]

    @property
    def phi(self):
        from km3pipe.math import phi_separg

        return phi_separg(self.dir_x, self.dir_y)

    @property
    def theta(self):
        from km3pipe.math import theta_separg

        return theta_separg(self.dir_z)

    @property
    def zenith(self):
        from km3pipe.math import neutrino_to_source_direction

        _, zen = neutrino_to_source_direction(self.phi, self.theta)
        return zen

    @property
    def azimuth(self):
        from km3pipe.math import neutrino_to_source_direction

        azi, _ = neutrino_to_source_direction(self.phi, self.theta)
        return azi

    @property
    def triggered_rows(self):
        if not hasattr(self, "triggered"):
            raise KeyError("Table has no 'triggered' column!")
        return self[self.triggered.astype(bool)]


class NDArray(np.ndarray):
    """Array with HDF5 metadata."""

    def __new__(cls, array, dtype=None, order=None, **kwargs):
        obj = np.asarray(array, dtype=dtype, order=order).view(cls)
        h5loc = kwargs.get("h5loc", "/misc")
        title = kwargs.get("title", "Unnamed NDArray")
        group_id = kwargs.get("group_id", None)
        obj.h5loc = h5loc
        obj.title = title
        obj.group_id = group_id
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.h5loc = getattr(obj, "h5loc", None)
        self.title = getattr(obj, "title", None)
        self.group_id = getattr(obj, "group_id", None)


class Vec3(object):
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

    def __add__(self, other):
        return Vec3(*np.add(self, other))

    def __radd__(self, other):
        return Vec3(*np.add(other, self))

    def __sub__(self, other):
        return Vec3(*np.subtract(self, other))

    def __rsub__(self, other):
        return Vec3(*np.subtract(other, self))

    def __mul__(self, other):
        return Vec3(*np.multiply(self, other))

    def __rmul__(self, other):
        return Vec3(*np.multiply(other, self))

    def __div__(self, other):
        return self.__truediv__(other)

    def __truediv__(self, other):
        return Vec3(*np.divide(self, other))

    def __array__(self, dtype=None):
        if dtype is not None:
            return np.array([self.x, self.y, self.z], dtype=dtype)
        else:
            return np.array([self.x, self.y, self.z])

    def __getitem__(self, index):
        return self.__array__()[index]



# Filename: tools.py
# pylint: disable=C0103
"""
Some unsorted, frequently used logic.

"""
import base64
import collections
from collections.abc import Mapping
from datetime import datetime, timedelta
import functools
import os
import re
import socket
import subprocess
import sys
import smtplib
import getpass

import numpy as np

__author__ = "Tamas Gal and Moritz Lotze"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = ["Konstantin Lepa <konstantin.lepa@gmail.com> for termcolor"]
__license__ = "MIT"
__maintainer__ = "Tamas Gal and Moritz Lotze"
__email__ = "tgal@km3net.de"
__status__ = "Development"

XROOTD_BASE = "root://ccxroot:1999"

File = collections.namedtuple("File", field_names=["path", "size"])


def ifiles(irods_path):
    """Return a list of File instances for the given iRODS path (recursively).

    The File instances offer `.path` and `.size` attributes.
    """
    if not iexists(irods_path):
        return []
    raw_output = subprocess.check_output("ils -lr {0}".format(irods_path), shell=True)
    filenames = {}
    base = irods_path
    for line in raw_output.splitlines():
        split_line = line.decode("ascii").strip().split()
        if len(split_line) == 1 and split_line[0].endswith(":"):
            base = split_line[0][:-1]  # remove trailing ':'
            continue
        if len(split_line) == 2 and split_line[0] == "C-":
            base = split_line[1]
            continue
        try:
            fsize = int(split_line[3])
            fname = split_line[6]
        except IndexError:
            import pdb

            pdb.set_trace()
        fpath = os.path.join(base, fname)
        filenames[fpath] = File(path=fpath, size=fsize)
    return list(filenames.values())


def iexists(irods_path):
    """Returns True of iRODS path exists, otherwise False"""
    try:
        subprocess.check_output(
            "ils {}".format(irods_path), shell=True, stderr=subprocess.PIPE,
        )
        return True
    except subprocess.CalledProcessError:
        return False


def isize(irods_path):
    """Returns the size in bytes of the most recent version of the file"""
    raw_output = subprocess.check_output(
        "ils -l {} | tail -n1 |awk '{{print $4}}'".format(irods_path), shell=True
    )
    try:
        return int(raw_output.decode("ascii").strip())
    except ValueError:
        raise IOError("File not found or an iRODS error occured.")


def xrdsize(xrootd_path):
    """Returns the size in bytes of the file"""
    base, path = re.search(r"(root://.*:[0-9]*)(/.*)", xrootd_path).groups()

    raw_output = subprocess.check_output(
        "xrdfs {} stat {} | grep Size | awk '{{print $2}}'".format(base, path),
        shell=True,
    )
    try:
        return int(raw_output.decode("ascii").strip())
    except ValueError:
        raise IOError("File not found or an xrootd error occured.")


def xrootd_path(det_id, run_id):
    """Return the xrootd path of a data file"""
    base = "root://ccxroot:1999//hpss/in2p3.fr/group/km3net/data/raw/sea"
    suffix = "KM3NeT_{:08d}/{}/KM3NeT_{:08d}_{:08d}.root".format(
        det_id, int(run_id / 1000), det_id, run_id
    )
    return os.path.join(base, suffix)


def token_urlsafe(nbytes=32):
    """Return a random URL-safe text string, in Base64 encoding.

    This is taken and slightly modified from the Python 3.6 stdlib.

    The string has *nbytes* random bytes.  If *nbytes* is ``None``
    or not supplied, a reasonable default is used.

    >>> token_urlsafe(16)  #doctest:+SKIP
    'Drmhze6EPcv0fN_81Bj-nA'

    """
    tok = os.urandom(nbytes)
    return base64.urlsafe_b64encode(tok).rstrip(b"=").decode("ascii")


def prettyln(text, fill="-", align="^", prefix="[ ", suffix=" ]", length=69):
    """Wrap `text` in a pretty line with maximum length."""
    text = "{prefix}{0}{suffix}".format(text, prefix=prefix, suffix=suffix)
    print(
        "{0:{fill}{align}{length}}".format(text, fill=fill, align=align, length=length)
    )


def irods_path(det_id, run_id):
    """Generate the iRODS filepath for given detector (O)ID and run ID"""
    data_path = "/in2p3/km3net/data/raw/sea"
    from km3pipe.db import DBManager

    if not isinstance(det_id, int):
        dts = DBManager().detectors
        det_id = int(dts[dts.OID == det_id].SERIALNUMBER.values[0])
    return data_path + "/KM3NeT_{0:08}/{2}/KM3NeT_{0:08}_{1:08}.root".format(
        det_id, run_id, run_id // 1000
    )


def unpack_nfirst(seq, nfirst):
    """Unpack the nfrist items from the list and return the rest.

    >>> a, b, c, rest = unpack_nfirst((1, 2, 3, 4, 5), 3)
    >>> a, b, c
    (1, 2, 3)
    >>> rest
    (4, 5)

    """
    iterator = iter(seq)
    for _ in range(nfirst):
        yield next(iterator, None)
    yield tuple(iterator)


def split(string, callback=None, sep=None):
    """Split the string and execute the callback function on each part.

    >>> string = "1 2 3 4"
    >>> parts = split(string, int)
    >>> parts
    [1, 2, 3, 4]

    """
    if callback is not None:
        return [callback(i) for i in string.split(sep)]
    else:
        return string.split(sep)


def namedtuple_with_defaults(typename, field_names, default_values=[]):
    """Create a namedtuple with default values

    >>> Node = namedtuple_with_defaults('Node', 'val left right')
    >>> Node()
    Node(val=None, left=None, right=None)
    >>> Node = namedtuple_with_defaults('Node', 'val left right', [1, 2, 3])
    >>> Node()
    Node(val=1, left=2, right=3)
    >>> Node = namedtuple_with_defaults('Node', 'val left right', {'right':7})
    >>> Node()
    Node(val=None, left=None, right=7)
    >>> Node(4)
    Node(val=4, left=None, right=7)
    """
    the_tuple = collections.namedtuple(typename, field_names)
    the_tuple.__new__.__defaults__ = (None,) * len(the_tuple._fields)
    if isinstance(default_values, Mapping):
        prototype = the_tuple(**default_values)
    else:
        prototype = the_tuple(*default_values)
    the_tuple.__new__.__defaults__ = tuple(prototype)
    return the_tuple


def remain_file_pointer(function):
    """Remain the file pointer position after calling the decorated function

    This decorator assumes that the last argument is the file handler.

    """

    def wrapper(*args, **kwargs):
        """Wrap the function and remain its parameters and return values"""
        file_obj = args[-1]
        old_position = file_obj.tell()
        return_value = function(*args, **kwargs)
        file_obj.seek(old_position, 0)
        return return_value

    return wrapper


def itervalues(d):
    return iter(d.values())


def iteritems(d):
    return iter(d.items())


def decamelise(text):
    """Convert CamelCase to lower_and_underscore."""
    s = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", text)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s).lower()


def camelise(text, capital_first=True):
    """Convert lower_underscore to CamelCase."""

    def camelcase():
        if not capital_first:
            yield str.lower
        while True:
            yield str.capitalize

    if istype(text, "unicode"):
        text = text.encode("utf8")
    c = camelcase()
    return "".join(next(c)(x) if x else "_" for x in text.split("_"))


ATTRIBUTES = dict(
    list(
        zip(
            ["bold", "dark", "", "underline", "blink", "", "reverse", "concealed"],
            list(range(1, 9)),
        )
    )
)
del ATTRIBUTES[""]

ATTRIBUTES_RE = r"\033\[(?:%s)m" % "|".join(["%d" % v for v in ATTRIBUTES.values()])

HIGHLIGHTS = dict(
    list(
        zip(
            [
                "on_grey",
                "on_red",
                "on_green",
                "on_yellow",
                "on_blue",
                "on_magenta",
                "on_cyan",
                "on_white",
            ],
            list(range(40, 48)),
        )
    )
)

HIGHLIGHTS_RE = r"\033\[(?:%s)m" % "|".join(["%d" % v for v in HIGHLIGHTS.values()])

COLORS = dict(
    list(
        zip(
            ["grey", "red", "green", "yellow", "blue", "magenta", "cyan", "white",],
            list(range(30, 38)),
        )
    )
)

COLORS_RE = r"\033\[(?:%s)m" % "|".join(["%d" % v for v in COLORS.values()])

RESET = r"\033[0m"
RESET_RE = r"\033\[0m"


def colored(text, color=None, on_color=None, attrs=None, ansi_code=None):
    """Colorize text, while stripping nested ANSI color sequences.

    Author:  Konstantin Lepa <konstantin.lepa@gmail.com> / termcolor

    Available text colors:
        red, green, yellow, blue, magenta, cyan, white.
    Available text highlights:
        on_red, on_green, on_yellow, on_blue, on_magenta, on_cyan, on_white.
    Available attributes:
        bold, dark, underline, blink, reverse, concealed.
    Example:
        colored('Hello, World!', 'red', 'on_grey', ['blue', 'blink'])
        colored('Hello, World!', 'green')
    """
    if os.getenv("ANSI_COLORS_DISABLED") is None:
        if ansi_code is not None:
            return "\033[38;5;{}m{}\033[0m".format(ansi_code, text)
        fmt_str = "\033[%dm%s"
        if color is not None:
            text = re.sub(COLORS_RE + "(.*?)" + RESET_RE, r"\1", text)
            text = fmt_str % (COLORS[color], text)
        if on_color is not None:
            text = re.sub(HIGHLIGHTS_RE + "(.*?)" + RESET_RE, r"\1", text)
            text = fmt_str % (HIGHLIGHTS[on_color], text)
        if attrs is not None:
            text = re.sub(ATTRIBUTES_RE + "(.*?)" + RESET_RE, r"\1", text)
            for attr in attrs:
                text = fmt_str % (ATTRIBUTES[attr], text)
        return text + RESET
    else:
        return text


def cprint(text, color=None, on_color=None, attrs=None):
    """Print colorize text.

    Author:  Konstantin Lepa <konstantin.lepa@gmail.com> / termcolor

    It accepts arguments of print function.
    """
    print((colored(text, color, on_color, attrs)))


def issorted(arr):
    """Check if array is sorted."""
    return np.all(np.diff(arr) >= 0)


def lstrip(text):
    """Remove leading whitespace from each line of a multiline string."""
    return "\n".join(l.lstrip() for l in text.lstrip().split("\n"))


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i : i + n]


def is_coherent(seq):
    """Find out if list of subsequent integers is complete.

    Adapted from https://stackoverflow.com/questions/18131741/python-find-out-whether-a-list-of-integers-is-coherent

    ```
    is_coherent([1, 2, 3, 4, 5]) -> True
    is_coherent([1,    3, 4, 5]) -> False
    ```
    """
    return np.array_equal(seq, range(seq[0], int(seq[-1] + 1)))


class AnyBar:
    """A lightweight interface to the AnyBar macOS app."""

    def __init__(self, port=1738, address="localhost"):
        self.port = port
        self.address = address
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def change(self, color):
        self.sock.sendto(color.encode("utf-8"), (self.address, self.port))


def zero_pad(m, n=1):
    """Pad a matrix with zeros, on all sides."""
    return np.pad(m, (n, n), mode="constant", constant_values=[0])


def istype(obj, typename):
    """Drop-in replacement for `isinstance` to avoid imports"""
    return type(obj).__name__ == typename


def isnotebook():
    """Check if running within a Jupyter notebook"""
    try:
        shell = get_ipython().__class__.__name__
        if shell == "ZMQInteractiveShell":
            return True  # Jupyter notebook or qtconsole
        elif shell == "TerminalInteractiveShell":
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False


def supports_color():
    """Checks if the terminal supports color."""
    if isnotebook():
        return True
    supported_platform = sys.platform != "win32" or "ANSICON" in os.environ
    is_a_tty = hasattr(sys.stdout, "isatty") and sys.stdout.isatty()

    if not supported_platform or not is_a_tty:
        return False

    return True


def get_jpp_version(via_command="JPrint -v"):
    """Return the Jpp version or None if not available."""
    try:
        out = subprocess.getoutput(via_command)
    except AttributeError:  # TODO: python 2.7
        try:
            out = subprocess.check_output(via_command.split(), stderr=subprocess.STDOUT)
        except OSError:
            return None

    for line in out.split("\n"):
        if line.startswith("version:"):
            jpp_version = line.split(":")[1].strip()
            return jpp_version

    return None


def timed_cache(**timed_cache_kwargs):
    """LRU cache decorator with timeout.

    Parameters
    ----------
    days: int
    seconds: int
    microseconds: int
    milliseconds: int
    minutes: int
    hours: int
    weeks: int
    maxsise: int [default: 128]
    typed: bool [default: False]
    """

    def _wrapper(f):
        maxsize = timed_cache_kwargs.pop("maxsize", 128)
        typed = timed_cache_kwargs.pop("typed", False)
        update_delta = timedelta(**timed_cache_kwargs)
        # nonlocal workaround to support Python 2
        # https://technotroph.wordpress.com/2012/10/01/python-closures-and-the-python-2-7-nonlocal-solution/
        d = {"next_update": datetime.utcnow() - update_delta}
        try:
            f = functools.lru_cache(maxsize=maxsize, typed=typed)(f)
        except AttributeError:
            print(
                "LRU caching is not available in Pyton 2.7, "
                "this will have no effect!"
            )
            pass

        @functools.wraps(f)
        def _wrapped(*args, **kwargs):
            now = datetime.utcnow()
            if now >= d["next_update"]:
                try:
                    f.cache_clear()
                except AttributeError:
                    pass
                d["next_update"] = now + update_delta
            return f(*args, **kwargs)

        return _wrapped

    return _wrapper


def sendmail(to, msg):
    """Send an email"""
    sender = "{}@{}".format(getpass.getuser(), socket.gethostname())
    s = smtplib.SMTP("localhost")
    s.sendmail(sender, to, msg)
    s.quit()



#!/usr/bin/env python
# vim: ts=4 sw=4 et

from io import BytesIO
import os

import numpy as np

import km3pipe as kp
from km3modules.plot import plot_dom_parameters
import km3pipe.style

km3pipe.style.use('km3pipe')

PLOTS_PATH = 'www/plots'
cal = kp.calib.Calibration(det_id=14)
detector = cal.detector


class MonitorRates(kp.Module):
    def configure(self):
        self.index = 0
        self.k40_2fold = {}
        self.rates = {}
        self.cuckoo = kp.time.Cuckoo(60, self.create_plot)
        self.n_slices = 0

    def process(self, blob):
        self.index += 1
        if self.index % 30:
            return blob

        data = blob['CHData']
        data_io = BytesIO(data)
        preamble = kp.io.daq.DAQPreamble(file_obj=data_io)    # noqa
        summaryslice = kp.io.daq.DAQSummaryslice(file_obj=data_io)
        for dom_id, rates in summaryslice.summary_frames.iteritems():
            du, dom, _ = detector.doms[dom_id]
            self.rates[(du, dom)] = np.sum(rates) / 1000

        self.cuckoo.msg()

        return blob

    def create_plot(self):
        print(self.__class__.__name__ + ": updating plot.")

        filename = os.path.join(PLOTS_PATH, 'dom_rates.png')
        plot_dom_parameters(
            self.rates,
            detector,
            filename,
            'rate [kHz]',
            "DOM Rates",
            vmin=190,
            vmax=230,
            cmap='coolwarm',
            missing='black',
            under='darkorchid',
            over='deeppink'
        )
        print("done")


pipe = kp.Pipeline()
pipe.attach(
    kp.io.ch.CHPump,
    host='127.0.0.1',
    port=5553,
    tags='IO_SUM',
    timeout=60 * 60 * 24 * 7,
    max_queue=2000
)
pipe.attach(kp.io.daq.DAQProcessor)
pipe.attach(MonitorRates)
pipe.drain()



import km3pipe as kp
import numpy as np
import pandas as pd

dt = np.dtype([("hi", int), ("there", float)])
dat = np.ones(5, dtype=dt)
df = pd.DataFrame.from_records(dat)


def foo(blob):
    blob["FooBar"] = df
    return blob


p = kp.Pipeline()
p.attach(foo)
p.attach(kp.io.HDF5Sink, filename="test.h5")
p.drain(10)



#!/usr/bin/env python
"""Tests for HDF5 stuff"""
from collections import OrderedDict, defaultdict
import tempfile
from os.path import join, dirname

import numpy as np
import tables as tb

from km3pipe import Blob, Module, Pipeline, Pump, version
from km3pipe.dataclasses import Table, NDArray
from km3pipe.io.hdf5 import (
    HDF5Pump,
    HDF5Sink,
    HDF5Header,
    convert_header_dict_to_table,
    FORMAT_VERSION,
)
from km3pipe.testing import TestCase, data_path


class TestH5Pump(TestCase):
    def setUp(self):
        self.fname = data_path(
            "hdf5/mcv5.40.mupage_10G.sirene.jterbr00006060.962.root.h5"
        )

    def test_init_sets_filename_if_no_keyword_arg_is_passed(self):
        p = HDF5Pump(filename=self.fname)
        self.assertEqual(self.fname, p.filename)
        p.finish()

    def test_standalone(self):
        pump = HDF5Pump(filename=self.fname)
        next(pump)
        pump.finish()

    def test_pipe(self):
        class Observer(Module):
            def configure(self):
                self.dump = defaultdict(list)

            def process(self, blob):
                for key, data in blob.items():
                    if key == "Header":
                        self.dump["headers"].append(data)
                    else:
                        self.dump[key].append(len(data))
                return blob

            def finish(self):
                return self.dump

        p = Pipeline()
        p.attach(HDF5Pump, filename=self.fname)
        p.attach(Observer)
        results = p.drain()["Observer"]
        self.assertListEqual(
            [147, 110, 70, 62, 59, 199, 130, 92, 296, 128], results["Hits"]
        )
        self.assertListEqual(
            [315, 164, 100, 111, 123, 527, 359, 117, 984, 263], results["McHits"]
        )
        self.assertListEqual([1, 1, 1, 1, 1, 3, 2, 1, 2, 1], results["McTracks"])

    def test_event_info_is_not_empty(self):
        self.fname = data_path("hdf5/test_event_info.h5")

        class Printer(Module):
            def process(self, blob):
                assert blob["EventInfo"].size != 0
                return blob

        p = Pipeline()
        p.attach(HDF5Pump, filename=self.fname)
        p.attach(Printer)
        p.drain()

    def test_event_info_has_correct_group_id(self):
        self.fname = data_path("hdf5/test_event_info.h5")

        class Printer(Module):
            def configure(self):
                self.index = 0

            def process(self, blob):
                assert blob["EventInfo"][0].group_id == self.index
                self.index += 1
                return blob

        p = Pipeline()
        p.attach(HDF5Pump, filename=self.fname)
        p.attach(Printer)
        p.drain()

    def test_get_blob(self):
        fname = data_path("hdf5/test_event_info.h5")
        pump = HDF5Pump(filename=fname)
        assert 44 == len(pump[0]["McTracks"])
        assert 3 == len(pump[1]["McTracks"])
        assert 179 == len(pump[2]["McTracks"])
        assert 55 == len(pump[3]["McTracks"])
        pump.finish()


class TestH5Sink(TestCase):
    def setUp(self):
        self.fname = data_path(
            "hdf5/mcv5.40.mupage_10G.sirene.jterbr00006060.962.root.h5"
        )
        self.fobj = tempfile.NamedTemporaryFile(delete=True)
        self.out = tb.open_file(
            self.fobj.name, "w", driver="H5FD_CORE", driver_core_backing_store=0
        )

    def tearDown(self):
        self.out.close()
        self.fobj.close()

    def test_pipe(self):
        p = Pipeline()
        p.attach(HDF5Pump, filename=self.fname)
        p.attach(HDF5Sink, h5file=self.out)
        p.drain()

    def test_h5info(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        class DummyPump(Pump):
            def process(self, blob):
                return Blob()

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(5)

        with tb.open_file(fname, "r") as h5file:
            assert version == h5file.root._v_attrs.km3pipe.decode()
            assert tb.__version__ == h5file.root._v_attrs.pytables.decode()
            assert FORMAT_VERSION == h5file.root._v_attrs.format_version

        fobj.close()

    def test_filtered_writing(self):
        fobjs = []
        for i in range(3):
            fobj = tempfile.NamedTemporaryFile(delete=True)
            fobjs.append(fobj)

        fobj_all = tempfile.NamedTemporaryFile(delete=True)

        class DummyPump(Pump):
            def configure(self):
                self.i = 0

            def process(self, blob):
                blob["A"] = Table({"a": self.i}, name="A", h5loc="tab_a")
                blob["B"] = Table({"b": self.i}, name="B", h5loc="tab_b")
                blob["C"] = Table({"c": self.i}, name="C", h5loc="tab_c")
                self.i += 1
                return blob

        keys = "ABC"

        pipe = Pipeline()
        pipe.attach(DummyPump)
        for fobj, key in zip(fobjs, keys):
            pipe.attach(HDF5Sink, filename=fobj.name, keys=[key])
        pipe.attach(HDF5Sink, filename=fobj_all.name)
        pipe.drain(5)

        for fobj, key in zip(fobjs, keys):
            with tb.File(fobj.name, "r") as f:
                assert "/tab_" + key.lower() in f
                for _key in set(keys) - set(key):
                    assert "/tab_" + _key.lower() not in f

        for key in keys:
            with tb.File(fobj_all.name, "r") as f:
                assert "/tab_" + key.lower() in f

        for fobj in fobjs:
            fobj.close()
        fobj_all.close()

    def test_filtered_writing_of_multiple_keys(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)

        class DummyPump(Pump):
            def configure(self):
                self.i = 0

            def process(self, blob):
                blob["A"] = Table({"a": self.i}, name="A", h5loc="tab_a")
                blob["B"] = Table({"b": self.i}, name="B", h5loc="tab_b")
                blob["C"] = Table({"c": self.i}, name="C", h5loc="tab_c")
                self.i += 1
                return blob

        keys = ["A", "B"]

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fobj.name, keys=keys)
        pipe.drain(5)

        with tb.File(fobj.name, "r") as f:
            assert "/tab_a" in f
            assert "/tab_b" in f
            assert "/tab_c" not in f

        fobj.close()

    def test_write_table_service(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)

        class Foo(Module):
            def prepare(self):
                self.services["write_table"](Table({"a": 1}, name="A", h5loc="tab_a"))

        pipe = Pipeline()
        pipe.attach(Foo)
        pipe.attach(HDF5Sink, filename=fobj.name)
        pipe.drain(5)

        with tb.File(fobj.name, "r") as f:
            assert "/tab_a" in f

        fobj.close()


class TestNDArrayHandling(TestCase):
    def test_writing_of_n_dim_arrays_with_defaults(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        arr = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])

        class DummyPump(Pump):
            def process(self, blob):
                blob["foo"] = NDArray(arr)
                return blob

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(3)

        with tb.File(fname) as f:
            foo = f.get_node("/misc")
            assert 3 == foo[0, 1, 0]
            assert 4 == foo[0, 1, 1]
            assert "Unnamed NDArray" == foo.title
            indices = f.get_node("/misc_indices")
            self.assertTupleEqual((0, 2, 4), tuple(indices.cols.index[:]))
            self.assertTupleEqual((2, 2, 2), tuple(indices.cols.n_items[:]))

        fobj.close()

    def test_writing_of_n_dim_arrays(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        arr = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])

        class DummyPump(Pump):
            def configure(self):
                self.index = 0

            def process(self, blob):
                blob["foo"] = NDArray(arr + self.index * 10, h5loc="/foo", title="Yep")
                self.index += 1
                return blob

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(3)

        with tb.File(fname) as f:
            foo = f.get_node("/foo")
            assert 3 == foo[0, 1, 0]
            assert 4 == foo[0, 1, 1]
            assert "Yep" == foo.title

        fobj.close()

    def test_writing_of_n_dim_arrays_in_nested_group(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        arr = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])

        class DummyPump(Pump):
            def configure(self):
                self.index = 0

            def process(self, blob):
                blob["foo"] = NDArray(arr + self.index * 10, h5loc="/foo/bar/baz")
                self.index += 1
                return blob

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(3)

        with tb.File(fname) as f:
            foo = f.get_node("/foo/bar/baz")
            print(foo)
            assert 3 == foo[0, 1, 0]
            assert 4 == foo[0, 1, 1]

        fobj.close()

    def test_writing_of_n_dim_arrays(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        arr = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])

        class DummyPump(Pump):
            def configure(self):
                self.index = 0

            def process(self, blob):
                blob["foo"] = NDArray(arr + self.index * 10, h5loc="/foo", title="Yep")
                self.index += 1
                return blob

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(3)

        with tb.File(fname) as f:
            foo = f.get_node("/foo")
            assert 3 == foo[0, 1, 0]
            assert 4 == foo[0, 1, 1]
            assert "Yep" == foo.title

        fobj.close()

    def test_reading_of_n_dim_arrays(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        arr = np.array([[[1, 2], [3, 4]], [[5, 6], [7, 8]]])

        class DummyPump(Pump):
            def configure(self):
                self.index = 0

            def process(self, blob):
                blob["Foo"] = NDArray(arr + self.index * 10, h5loc="/foo", title="Yep")
                self.index += 1
                return blob

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(3)

        class Observer(Module):
            def configure(self):
                self.index = 0

            def process(self, blob):
                assert "Foo" in blob
                foo = blob["Foo"]
                print(self.index)
                assert self.index * 10 + 1 == foo[0, 0, 0]
                assert self.index * 10 + 8 == foo[1, 1, 1]
                assert self.index * 10 + 3 == foo[0, 1, 0]
                assert self.index * 10 + 6 == foo[1, 0, 1]
                self.index += 1
                return blob

        pipe = Pipeline()
        pipe.attach(HDF5Pump, filename=fname)
        pipe.attach(Observer)
        pipe.drain()

        fobj.close()


class TestH5SinkConsistency(TestCase):
    def test_h5_consistency_for_tables_without_group_id(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        class DummyPump(Pump):
            def configure(self):
                self.count = 0

            def process(self, blob):
                self.count += 10
                tab = Table({"a": self.count, "b": 1}, h5loc="tab")
                return Blob({"tab": tab})

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(5)

        with tb.File(fname) as f:
            a = f.get_node("/tab")[:]["a"]
            b = f.get_node("/tab")[:]["b"]
            group_id = f.get_node("/tab")[:]["group_id"]
        assert np.allclose([10, 20, 30, 40, 50], a)
        assert np.allclose([1, 1, 1, 1, 1], b)
        assert np.allclose([0, 1, 2, 3, 4], group_id)
        fobj.close()

    def test_h5_consistency_for_tables_without_group_id_and_multiple_keys(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        class DummyPump(Pump):
            def configure(self):
                self.count = 0

            def process(self, blob):
                self.count += 10
                tab1 = Table({"a": self.count, "b": 1}, h5loc="tab1")
                tab2 = Table({"c": self.count + 1, "d": 2}, h5loc="tab2")
                return Blob({"tab1": tab1, "tab2": tab2})

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(5)

        with tb.File(fname) as f:
            a = f.get_node("/tab1")[:]["a"]
            b = f.get_node("/tab1")[:]["b"]
            c = f.get_node("/tab2")[:]["c"]
            d = f.get_node("/tab2")[:]["d"]
            group_id_1 = f.get_node("/tab1")[:]["group_id"]
            group_id_2 = f.get_node("/tab1")[:]["group_id"]
        assert np.allclose([10, 20, 30, 40, 50], a)
        assert np.allclose([1, 1, 1, 1, 1], b)
        assert np.allclose([0, 1, 2, 3, 4], group_id_1)
        assert np.allclose([11, 21, 31, 41, 51], c)
        assert np.allclose([2, 2, 2, 2, 2], d)
        assert np.allclose([0, 1, 2, 3, 4], group_id_2)
        fobj.close()

    def test_h5_consistency_for_tables_with_custom_group_id(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        class DummyPump(Pump):
            def process(self, blob):
                tab = Table({"group_id": 2}, h5loc="tab")
                return Blob({"tab": tab})

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(5)

        with tb.File(fname) as f:
            group_id = f.get_node("/tab")[:]["group_id"]

        assert np.allclose([2, 2, 2, 2, 2], group_id)

        fobj.close()

    def test_h5_singletons(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        class DummyPump(Pump):
            def process(self, blob):
                tab = Table({"a": 2}, h5loc="tab", h5singleton=True)
                return Blob({"tab": tab})

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(5)

        with tb.File(fname) as f:
            a = f.get_node("/tab")[:]["a"]

        assert len(a) == 1

        fobj.close()

    def test_h5_singletons_reading(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        class DummyPump(Pump):
            def process(self, blob):
                tab = Table({"a": 2}, h5loc="tab", h5singleton=True)
                return Blob({"Tab": tab})

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(5)

        class Observer(Module):
            def process(self, blob):
                print(blob)
                assert "Tab" in blob
                print(blob["Tab"])
                assert len(blob["Tab"]) == 1
                assert blob["Tab"].a[0] == 2
                return blob

        pipe = Pipeline()
        pipe.attach(HDF5Pump, filename=fname)
        pipe.attach(Observer)
        pipe.drain()

        fobj.close()


class TestHDF5PumpConsistency(TestCase):
    def test_hdf5_readout(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        class DummyPump(Pump):
            def configure(self):
                self.count = 0

            def process(self, blob):
                self.count += 1
                tab = Table({"a": self.count * 10, "b": 1}, h5loc="tab")
                tab2 = Table({"a": np.arange(self.count)}, h5loc="tab2")
                blob["Tab"] = tab
                blob["Tab2"] = tab2
                return blob

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(5)

        class BlobTester(Module):
            def configure(self):
                self.index = 0

            def process(self, blob):
                self.index += 1
                assert "GroupInfo" in blob
                assert "Tab" in blob
                print(self.index)
                print(blob["Tab"])
                print(blob["Tab"]["a"])
                assert self.index - 1 == blob["GroupInfo"].group_id
                assert self.index * 10 == blob["Tab"]["a"]
                assert 1 == blob["Tab"]["b"] == 1
                assert np.allclose(np.arange(self.index), blob["Tab2"]["a"])
                return blob

        pipe = Pipeline()
        pipe.attach(HDF5Pump, filename=fname)
        pipe.attach(BlobTester)
        pipe.drain()

        fobj.close()

    def test_hdf5_readout_split_tables(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        class DummyPump(Pump):
            def configure(self):
                self.count = 0

            def process(self, blob):
                self.count += 1
                tab = Table({"a": self.count * 10, "b": 1}, h5loc="/tab", split_h5=True)
                blob["Tab"] = tab
                return blob

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(5)

        class BlobTester(Module):
            def configure(self):
                self.index = 0

            def process(self, blob):
                self.index += 1
                assert "GroupInfo" in blob
                assert "Tab" in blob
                assert self.index - 1 == blob["GroupInfo"].group_id
                assert self.index * 10 == blob["Tab"]["a"]
                assert 1 == blob["Tab"]["b"]
                return blob

        pipe = Pipeline()
        pipe.attach(HDF5Pump, filename=fname)
        pipe.attach(BlobTester)
        pipe.drain()

        fobj.close()

    def test_hdf5_readout_split_tables_in_same_group(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        class DummyPump(Pump):
            def configure(self):
                self.count = 0

            def process(self, blob):
                self.count += 1
                tab_a = Table(
                    {"a": self.count * 10,}, h5loc="/tabs/tab_a", split_h5=True
                )
                tab_b = Table(
                    {"b": self.count * 100,}, h5loc="/tabs/tab_b", split_h5=True
                )
                blob["TabA"] = tab_a
                blob["TabB"] = tab_b
                return blob

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(5)

        class BlobTester(Module):
            def configure(self):
                self.index = 0

            def process(self, blob):
                self.index += 1
                assert "GroupInfo" in blob
                assert "TabA" in blob
                assert "TabB" in blob
                assert self.index - 1 == blob["GroupInfo"].group_id
                assert self.index * 10 == blob["TabA"]["a"]
                assert self.index * 100 == blob["TabB"]["b"]
                return blob

        pipe = Pipeline()
        pipe.attach(HDF5Pump, filename=fname)
        pipe.attach(BlobTester)
        pipe.drain()

        fobj.close()

    def test_sparse_table(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        class Dummy(Module):
            def configure(self):
                self.i = 0

            def process(self, blob):
                self.i += 1

                if self.i == 5:
                    blob["Tab"] = Table({"a": 23}, h5loc="/tab")
                return blob

        pipe = Pipeline()
        pipe.attach(Dummy)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(10)

        class Observer(Module):
            def configure(self):
                self.i = 0

            def process(self, blob):
                self.i += 1

                if self.i == 5:
                    assert 23 == blob["Tab"].a[0]
                else:
                    assert "Tab" not in blob

                return blob

        pipe = Pipeline()
        pipe.attach(HDF5Pump, filename=fname)
        pipe.attach(Observer)
        pipe.drain()

    # TODO: This will fail due to ugly indexing of NDArrays
    # def test_sparse_ndarray(self):
    #     fobj = tempfile.NamedTemporaryFile(delete=True)
    #     fname = fobj.name
    #
    #     class Dummy(Module):
    #         def configure(self):
    #             self.i = 0
    #
    #         def process(self, blob):
    #             self.i += 1
    #
    #             if self.i == 5:
    #                 blob['Arr'] = NDArray([1, 2, 3], h5loc='/arr')
    #             return blob
    #
    #     pipe = Pipeline()
    #     pipe.attach(Dummy)
    #     pipe.attach(HDF5Sink, filename=fname)
    #     pipe.drain(10)
    #
    #     class Observer(Module):
    #         def configure(self):
    #             self.i = 0
    #
    #         def process(self, blob):
    #             self.i += 1
    #
    #             print(blob)
    #             if self.i == 5:
    #                 assert 6 == np.sum(blob['Arr'])
    #             # else:
    #             #     assert 'Arr' not in blob
    #
    #             return blob
    #
    #     pipe = Pipeline()
    #     pipe.attach(HDF5Pump, filename=fname)
    #     pipe.attach(Observer)
    #     pipe.drain()


class TestHDF5Shuffle(TestCase):
    def test_shuffle_without_reset_index(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        class DummyPump(Pump):
            def configure(self):
                self.i = 0

            def process(self, blob):
                blob["Tab"] = Table({"a": self.i}, h5loc="/tab")
                blob["SplitTab"] = Table(
                    {"b": self.i}, h5loc="/split_tab", split_h5=True
                )
                blob["Arr"] = NDArray(np.arange(self.i + 1), h5loc="/arr")
                self.i += 1
                return blob

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(5)

        shuffled_group_ids = [2, 1, 0, 3, 4]

        def shuffle(x):
            for i in range(len(x)):
                x[i] = shuffled_group_ids[i]

        class Observer(Module):
            def configure(self):
                self.group_ids_tab = []
                self.group_ids_split_tab = []
                self.group_ids_arr = []
                self.a = []
                self.b = []
                self.arr_len = []

            def process(self, blob):
                group_id_tab = blob["Tab"].group_id[0]
                group_id_split_tab = blob["SplitTab"].group_id[0]
                group_id_arr = blob["Arr"].group_id
                assert blob["GroupInfo"].group_id[0] == group_id_tab
                assert blob["GroupInfo"].group_id[0] == group_id_split_tab
                assert blob["GroupInfo"].group_id[0] == group_id_arr
                self.group_ids_tab.append(blob["Tab"].group_id[0])
                self.group_ids_split_tab.append(blob["SplitTab"].group_id[0])
                self.group_ids_arr.append(blob["Arr"].group_id)
                self.a.append(blob["Tab"].a[0])
                self.b.append(blob["SplitTab"].b[0])
                self.arr_len.append(len(blob["Arr"]) - 1)
                return blob

            def finish(self):
                return {
                    "group_ids_tab": self.group_ids_tab,
                    "group_ids_split_tab": self.group_ids_split_tab,
                    "group_ids_arr": self.group_ids_arr,
                    "a": self.a,
                    "b": self.b,
                    "arr_len": self.arr_len,
                }

        pipe = Pipeline()
        pipe.attach(
            HDF5Pump,
            filename=fname,
            shuffle=True,
            shuffle_function=shuffle,
            reset_index=False,
        )
        pipe.attach(Observer)
        results = pipe.drain()

        self.assertListEqual(results["Observer"]["group_ids_tab"], shuffled_group_ids)
        self.assertListEqual(
            results["Observer"]["group_ids_split_tab"], shuffled_group_ids
        )
        self.assertListEqual(results["Observer"]["group_ids_arr"], shuffled_group_ids)
        self.assertListEqual(results["Observer"]["a"], shuffled_group_ids)
        self.assertListEqual(results["Observer"]["b"], shuffled_group_ids)
        # a small hack: we store the length of the array in 'b', which is
        # then equal to the shuffled group IDs (since those were generated
        # using the group_id
        self.assertListEqual(results["Observer"]["arr_len"], shuffled_group_ids)

        fobj.close()

    def test_shuffle_with_reset_index(self):
        fobj = tempfile.NamedTemporaryFile(delete=True)
        fname = fobj.name

        class DummyPump(Pump):
            def configure(self):
                self.i = 0

            def process(self, blob):
                blob["Tab"] = Table({"a": self.i}, h5loc="/tab")
                blob["SplitTab"] = Table(
                    {"b": self.i}, h5loc="/split_tab", split_h5=True
                )
                blob["Arr"] = NDArray(np.arange(self.i + 1), h5loc="/arr")
                self.i += 1
                return blob

        pipe = Pipeline()
        pipe.attach(DummyPump)
        pipe.attach(HDF5Sink, filename=fname)
        pipe.drain(5)

        shuffled_group_ids = [2, 1, 0, 3, 4]

        def shuffle(x):
            for i in range(len(x)):
                x[i] = shuffled_group_ids[i]

        class Observer(Module):
            def configure(self):
                self.group_ids_tab = []
                self.group_ids_split_tab = []
                self.group_ids_arr = []
                self.a = []
                self.b = []
                self.arr_len = []

            def process(self, blob):
                group_id_tab = blob["Tab"].group_id[0]
                group_id_split_tab = blob["SplitTab"].group_id[0]
                group_id_arr = blob["Arr"].group_id
                assert blob["GroupInfo"].group_id[0] == group_id_tab
                assert blob["GroupInfo"].group_id[0] == group_id_split_tab
                assert blob["GroupInfo"].group_id[0] == group_id_arr
                self.group_ids_tab.append(blob["Tab"].group_id[0])
                self.group_ids_split_tab.append(blob["SplitTab"].group_id[0])
                self.group_ids_arr.append(blob["Arr"].group_id)
                self.a.append(blob["Tab"].a[0])
                self.b.append(blob["SplitTab"].b[0])
                self.arr_len.append(len(blob["Arr"]) - 1)
                return blob

            def finish(self):
                return {
                    "group_ids_tab": self.group_ids_tab,
                    "group_ids_split_tab": self.group_ids_split_tab,
                    "group_ids_arr": self.group_ids_arr,
                    "a": self.a,
                    "b": self.b,
                    "arr_len": self.arr_len,
                }

        pipe = Pipeline()
        pipe.attach(
            HDF5Pump,
            filename=fname,
            shuffle=True,
            shuffle_function=shuffle,
            reset_index=True,
        )
        pipe.attach(Observer)
        results = pipe.drain()

        self.assertListEqual(results["Observer"]["group_ids_tab"], [0, 1, 2, 3, 4])
        self.assertListEqual(
            results["Observer"]["group_ids_split_tab"], [0, 1, 2, 3, 4]
        )
        self.assertListEqual(results["Observer"]["group_ids_arr"], [0, 1, 2, 3, 4])
        self.assertListEqual(results["Observer"]["a"], shuffled_group_ids)
        self.assertListEqual(results["Observer"]["b"], shuffled_group_ids)
        # a small hack: we store the length of the array in 'b', which is
        # then equal to the shuffled group IDs (since those were generated
        # using the group_id
        self.assertListEqual(results["Observer"]["arr_len"], shuffled_group_ids)

        fobj.close()


class TestHDF5Header(TestCase):
    def setUp(self):
        # self.hdict = OrderedDict([
        # # yapf crushes the formatting, never mind...
        # # we use OrderedDict here to ensure the correct ordering
        #     ("param_a", OrderedDict([("field_a_1", "1"), ("field_a_2", "2")])),
        #     ("param_b", OrderedDict([("field_b_1", "a")])),
        #     ("param_c", OrderedDict([("field_c_1", 23)])),
        #     (
        #         "param_d",
        #         OrderedDict([("param_d_0", 1), ("param_d_1", 2),
        #                      ("param_d_2", 3)])
        #     )
        # ])
        self.hdict = {
            "param_a": {"field_a_1": "1", "field_a_2": "2"},
            "param_b": {"field_b_1": "a"},
            "param_c": {"field_c_1": 23},
            "param_d": {"param_d_0": 1, "param_d_1": 2, "param_d_2": 3},
            "param_e": {"param_e_2": 3, "param_e_0": 1, "param_e_1": 2},
        }

    def test_init(self):
        HDF5Header({})

    def test_header(self):
        header = HDF5Header(self.hdict)
        assert "1" == header.param_a.field_a_1
        assert "2" == header.param_a.field_a_2
        assert "a" == header.param_b.field_b_1
        assert 23 == header.param_c.field_c_1

    def test_header_with_vectors(self):
        header = HDF5Header(self.hdict)
        self.assertTupleEqual((1, 2, 3), header.param_d)

    def test_header_with_scrumbled_vectors(self):
        header = HDF5Header(self.hdict)
        self.assertTupleEqual((1, 2, 3), header.param_e)

    # def test_header_with_scalars(self):
    #     header = HDF5Header(self.hdict)
    #     assert 4 == header.param_e
    #     assert 5.6 == header.param_f
    #
    # def test_scientific_notation(self):
    #     header = HDF5Header(self.hdict)
    #     assert 7e+08 == header.param_g

    def test_header_from_table(self):
        table = convert_header_dict_to_table(self.hdict)
        header = HDF5Header.from_table(table)
        print(header)
        assert 1.0 == header.param_a.field_a_1
        assert 2.0 == header.param_a.field_a_2
        assert "a" == header.param_b.field_b_1
        assert 23 == header.param_c.field_c_1
        self.assertTupleEqual((1, 2, 3), header.param_d)

    def test_header_from_hdf5_file(self):
        header = HDF5Header.from_hdf5(data_path("hdf5/raw_header.h5"))
        assert "MUSIC" == header.propag[0]
        assert "seawater" == header.propag[1]
        assert 3450 == header.seabottom[0]
        self.assertAlmostEqual(12.1, header.livetime.numberOfSeconds, places=3)
        self.assertAlmostEqual(0.09, header.livetime.errorOfSeconds, places=3)
        assert 0 == header.coord_origin.x
        assert 0 == header.coord_origin.y
        assert 0 == header.coord_origin.z
        self.assertTupleEqual((0, 0, 0), header.coord_origin)

    def test_header_from_table_with_bytes(self):
        table = Table(
            {
                "dtype": [b"f4 a2", b"f4"],
                "field_names": [b"a b", b"c"],
                "field_values": [b"1.2 ab", b"3.4"],
                "parameter": [b"foo", b"bar"],
            }
        )
        header = HDF5Header.from_aanet(table)
        self.assertAlmostEqual(1.2, header.foo.a, places=2)
        assert "ab" == header.foo.b
        self.assertAlmostEqual(3.4, header.bar.c, places=2)


class TestConvertHeaderDictToTable(TestCase):
    def setUp(self):
        hdict = {
            "param_a": {"field_a_1": "1", "field_a_2": "2"},
            "param_b": {"field_b_1": "a"},
            "param_c": {"field_c_1": 1},
        }
        self.tab = convert_header_dict_to_table(hdict)

    def test_length(self):
        assert 3 == len(self.tab)

    def test_values(self):
        tab = self.tab

        index_a = tab.parameter.tolist().index("param_a")
        index_b = tab.parameter.tolist().index("param_b")

        assert "param_a" == tab.parameter[index_a]
        assert "field_a_1" in tab.field_names[index_a]
        assert "field_a_2" in tab.field_names[index_a]
        if "field_a_1 field_a_2" == tab.field_names[index_a]:
            assert "1 2" == tab.field_values[index_a]
        else:
            assert "2 1" == tab.field_values[index_a]
        assert "f4 f4" == tab["dtype"][index_a]

        assert "param_b" == tab.parameter[index_b]
        assert "field_b_1" == tab.field_names[index_b]
        assert "a" == tab.field_values[index_b]
        assert "a1" == tab["dtype"][index_b]

    def test_values_are_converted_to_str(self):
        index_c = self.tab.parameter.tolist().index("param_c")
        assert "param_c" == self.tab.parameter[index_c]
        assert "1" == self.tab.field_values[index_c]

    def test_conversion_returns_none_for_empty_dict(self):
        assert None is convert_header_dict_to_table(None)
        assert None is convert_header_dict_to_table({})



__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class UI(object):
    """Represents the settings for the UI."""

    fg = "light gray"
    bg = "black"

    palette = [
        ("default", fg, bg),
        ("highlight", fg + ",standout", bg),
        ("header", "white", "dark cyan"),
        ("footer", "light gray", "dark blue"),
        ("body", "dark cyan", "", "standout"),
        ("focus", "dark red", "", "standout"),
        ("head", "light red", "black"),
        ("blob", "yellow", "dark cyan"),
        ("blob_selected", "dark cyan", "yellow"),
        ("blob_scale", "dark cyan", "black"),
    ]

    keys = {
        "select": ("return", "enter"),
        "inspect": ("x", "X"),
        "escape": ("esc", "q", "Q"),
        "left": ("left", "h"),
        "right": ("right", "l"),
        "up": ("up", "k"),
        "down": ("down", "j"),
        "home": ("0", "^"),
        "end": ("$",),
        "goto": ("g", "G"),
        "help": ("?",),
    }



#!/usr/bin/env python
"""
=================
Slackbot for ELOG
=================

"""
import os
import re

import pyinotify

from pyslack import SlackClient  # pip install pyslack-real

wm = pyinotify.WatchManager()
mask = pyinotify.IN_CLOSE_WRITE

URL = "http://elog.km3net.de"
# LOGBOOK_PATH = '/home/elog/elog'
LOGBOOK_PATH = "/usr/local/elog/logbooks"
BOTNAME = "ELOG"
DEFAULT_DESTINATION = "#elog"
DESTINATIONS = {
    "Operations_IT": "#operations_it",
    "Operations_FR": "#operations_fr",
    #    'Qualification': '#elog',
    #    'DOM_Integration': '#elog',
    #    'DU_Integration': '#elog',
    #    'DAQ_Readout': '#elog',
    #    'Electronics': '#elog',
    #    'Analysis': '#elog',
    "Computing and Software": "#software",
}

slack = SlackClient("YOUR_SLACK_API_TOKEN_HERE")


class ElogEntry(object):
    def __init__(self, logbook, msg_id=None):
        self.logbook = logbook
        self.id = msg_id
        self.header = {}
        self.content = ""

    @property
    def author(self):
        return self._lookup("Author")

    @property
    def type(self):
        return self._lookup("Type")

    @property
    def subject(self):
        return self._lookup("Subject")

    @property
    def url(self):
        escaped = re.sub(r"[\s_]+", "+", self.logbook)
        return URL + "/" + escaped + "/" + str(self.id)

    def _lookup(self, key):
        try:
            return self.header[key]
        except KeyError:
            return "Unknown"

    def __repr__(self):
        return (
            "ELOG entry from *{0.author}* in *{0.logbook}* (_{0.type}_):\n"
            "*{0.subject}*\n{1}...\n<{0.url}>".format(self, self.content[:150])
        )


class ElogEntryBundle(object):
    def __init__(self, pathname):
        self.logbook = dirname(pathname)

        with open(pathname) as input_file:
            self.entries = self._parse_entries(input_file)

    @property
    def newest_entry(self):
        return self.entries[-1]

    def _parse_entries(self, input_file, msg_id=None):
        if msg_id is None:
            line = input_file.readline()
            if not line.startswith("$@MID@$:"):
                raise ValueError("Not an ELOG-entry file!")
            msg_id = int(line.split(":")[1])

        entry = ElogEntry(self.logbook, msg_id)
        for line in input_file:
            if line.startswith(40 * "="):
                break
            parameter, value = re.findall(r"([^:.]*): (.*)", line)[0]
            entry.header[parameter] = value
        for line in input_file:
            if line.startswith("$@MID@$:"):
                msg_id = int(line.split(":")[1])
                return [entry] + self._parse_entries(input_file, msg_id)
            else:
                entry.content += line
        return [entry]


class EventHandler(pyinotify.ProcessEvent):
    def my_init(self):
        self.logged_ids = []

    def process_IN_CLOSE_WRITE(self, event):
        if not self._is_valid_filetype(event.pathname):
            print("Not a valid filetype: {0}".format(event.pathname))
            return

        try:
            elog_bundle = ElogEntryBundle(event.pathname)
            elog_entry = elog_bundle.newest_entry
        except ValueError:
            print("Could not parse '{0}'. Ignoring...".format(event.pathname))
            return

        try:
            destination = DESTINATIONS[elog_entry.logbook]
        except KeyError:
            print(
                "No destination for logbook '{0}'. Using only default...".format(
                    elog_entry.logbook
                )
            )
            destination = None
        finally:
            if elog_entry.id in self.logged_ids:
                return  # For now skip, since it often duplicates!
                pre = "Updated "
            else:
                pre = ""
                self.logged_ids.append(elog_entry.id)
            message = pre + str(elog_entry)
            print(42 * "#")
            print("Filename: " + event.name)
            print("Elog Entry ID: " + str(elog_entry.id))
            print(42 * "-")
            print(message)
            print(42 * "/")
            if destination:
                slack.chat_post_message(destination, message, username=BOTNAME)
            # else:
            slack.chat_post_message(DEFAULT_DESTINATION, message, username=BOTNAME)

    def _is_valid_filetype(self, path):
        return path.endswith(".log")


def dirname(filepath):
    return os.path.basename(os.path.dirname(filepath))


handler = EventHandler()
notifier = pyinotify.Notifier(wm, handler)
wdd = wm.add_watch(LOGBOOK_PATH, mask, rec=True)

notifier.loop()



# -*- coding: utf-8 -*-
"""

====
Cone
====

Sparse Cone
"""

import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D  # noqa

from km3pipe.math import SparseCone
import km3pipe.style

km3pipe.style.use("moritz")

spike = [0, 0, 1]
bottom = [0, 0, 0]
angle = np.pi / 4
n_angles = 20
cone = SparseCone(spike, bottom, angle)
circ_samp = cone.sample_circle(n_angles=n_angles)
axis_samp = cone.sample_axis
samp = cone.sample(n_angles)

samp = np.array(samp)

##############################################################################
# plot the same in 3D because why not?

plt.clf()
fig = plt.figure()
ax = fig.add_subplot(111, projection="3d")
ax.scatter(samp[:, 0], samp[:, 1], samp[:, 2], "yo")
plt.show()



"""Read & dump events through the CH Pump.
"""
from itertools import combinations

import numpy as np

from km3pipe import Pipeline, Module
from km3pipe.io.ch import CHPump
from km3pipe.io.daq import TimesliceParser


def printer(blob):
    print(blob["TimesliceFrames"][808447091])
    return blob


def mongincidence(times, tdcs, tmax=20):
    coincidences = []
    cur_t = 0
    las_t = 0
    for t_idx, t in enumerate(times):
        cur_t = t
        diff = cur_t - las_t
        if diff <= tmax and t_idx > 0 and tdcs[t_idx - 1] != tdcs[t_idx]:
            coincidences.append(((tdcs[t_idx - 1], tdcs[t_idx]), diff))
        las_t = cur_t
    return coincidences


class CoincidenceFinder(Module):
    def configure(self):
        self.m = np.zeros(shape=(465, 41))

    def process(self, blob):
        hits = blob["TimesliceFrames"][808447091]
        hits.sort(key=lambda x: x[1])
        coinces = mongincidence([t for (_, t, _) in hits], [t for (t, _, _) in hits])

        combs = list(combinations(range(31), 2))
        for pmt_pair, t in coinces:
            if pmt_pair[0] > pmt_pair[1]:
                pmt_pair = (pmt_pair[1], pmt_pair[0])
                t = -t
            self.m[combs.index(pmt_pair), t + 20] += 1

        print(self.m)

        return blob

    def finish(self):
        np.savetxt("coincidences.txt", self.m)


class Dumper(Module):
    def configure(self):
        self.counter = 0

    def process(self, blob):
        print("New blob:")
        print(blob["CHPrefix"])
        if "CHData" in blob:
            tag = str(blob["CHPrefix"].tag)
            data = blob["CHData"]
            self.dump(data, tag)
        return blob

    def dump(self, data, tag):
        with open("{0}-{1:06}.dat".format(tag, self.counter), "bw") as f:
            self.counter += 1
            f.write(data)


pipe = Pipeline()
pipe.attach(
    CHPump,
    host="172.16.65.58",
    port=5553,
    tags="IO_TSL",
    timeout=60 * 60 * 24,
    max_queue=42,
)
# pipe.attach(Dumper)
pipe.attach(TimesliceParser)
# pipe.attach(printer)
pipe.attach(CoincidenceFinder)
pipe.drain()



#!/usr/bin/env python
"""
Usage:
    kp_jfit_readout.py FILENAME
"""

from docopt import docopt

from km3pipe.io.hdf5 import HDF5Sink
from km3pipe.io.jpp import FitPump
from km3pipe import Pipeline


def print_fits(blob):
    fits = blob["JFit"]
    print(fits[:10])


if __name__ == "__main__":
    args = docopt(__doc__)
    fname = args["FILENAME"]

    pipe = Pipeline()
    pipe.attach(FitPump, filename=fname)
    pipe.attach(print_fits)
    pipe.attach(HDF5Sink, filename=fname + ".h5")
    pipe.drain(1)



#!/usr/bin/env python
# Filename: qrunprocessor.py
# Author: Tamas Gal <tgal@km3net.de>
"""
=========================================
Create a summary for a given list of runs
=========================================

Use this batch runner to process a given list of run numbers with a
script, which takes a `-o` to create a summary file, which has the name
of the processed file + a given suffix.
The actual call will look like this, with a bit of copy safeness:

    /abs/path/to/SCRIPT run.root -o /abs/path/to/OUTPUT_PATH+SUFFIX

Before constructing the job scripts, the ``OUTPUT_PATH`` will be traversed
to find files which have already been converted to avoid multiple conversions.

.. code-block:: console

    Usage:
        qrunprocessor [options] DET_ID RUN_LIST OUTPUT_PATH SCRIPT
        qrunprocessor (-h | --help)

    Options:
        DET_ID         Detector ID (e.g. 29).
        RUN_LIST       Path to the file containing the space separated run IDs.
        OUTPUT_PATH    Path to store the individual summary files.
        SCRIPT         The script to fire up.
        -s SUFFIX      The suffix, appended by SCRIPT [default: .summary.h5].
        -n N_FILES     Number of files to process per job [default: 10].
        -e ET          Estimated walltime per file in minutes [default: 15].
        -f FSIZE       Estimated filesystem size for a job [default: 12G].
        -m VMEM        Estimated vmem for a job [default: 8G].
        -j JOBNAME     The name of the submitted jobs [default: qrunprocessor].
        -l LOG_PATH    Path of the job log files [default: qlogs].
        -v PYTHONVENV  Path to the Python virtual env.
        -c CLUSTER     Cluster to run on (in2p3, woody, ...) [default: in2p3].
        -q             Dryrun: don't submit jobs, just print the job script.
        -h --help      Show this screen.

"""
__author__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__version__ = "1.0"


def main():
    from docopt import docopt

    args = docopt(__doc__, version=__version__)

    from glob import glob
    import os
    from os.path import basename, join, abspath
    import pathlib
    import time
    import km3pipe as kp
    from km3pipe.tools import chunks, iexists

    try:
        from tqdm import tqdm
    except ImportError:

        def tqdm(x):
            return x

    cprint = kp.logger.get_printer("qrunprocessor")
    log = kp.logger.get_logger("qrunprocessor")

    RUN_LIST = os.path.abspath(args["RUN_LIST"])
    OUTPUT_PATH = os.path.abspath(args["OUTPUT_PATH"])
    SCRIPT = os.path.abspath(args["SCRIPT"])
    SUFFIX = args["-s"]
    DET_ID = int(args["DET_ID"])
    ET_PER_FILE = int(args["-e"]) * 60  # [s]
    FILES_PER_JOB = int(args["-n"])
    FSIZE = args["-f"]
    VMEM = args["-m"]
    LOG_PATH = os.path.abspath(args["-l"])
    JOB_NAME = args["-j"]
    DRYRUN = args["-q"]
    PYTHONVENV = os.path.abspath(args["-v"])
    CLUSTER = args["-c"]

    pathlib.Path(OUTPUT_PATH).mkdir(parents=True, exist_ok=True)

    with open(RUN_LIST, "r") as fobj:
        run_numbers = [int(run) for run in fobj.read().split()]

    xrootd_files = []
    for run in run_numbers:
        xrootd_path = kp.tools.xrootd_path(DET_ID, run)
        xrootd_files.append(xrootd_path)

    processed_files = [
        basename(f) for f in glob(join(OUTPUT_PATH, "*{}".format(SUFFIX)))
    ]

    rem_files = []
    for xrootd_file in xrootd_files:
        if basename(xrootd_file) + SUFFIX not in processed_files:
            rem_files.append(xrootd_file)

    cprint(
        "{} runs in total, {} already processed.".format(
            len(xrootd_files), len(processed_files)
        )
    )
    cprint(f"Proceeding with the remaining {len(rem_files)} files.")

    s = kp.shell.Script()

    for job_id, file_chunk in enumerate(chunks(rem_files, FILES_PER_JOB)):
        n_files = len(file_chunk)
        s.add(f"echo Creating run summary for {n_files} files")
        s.add("cd $TMPDIR; mkdir -p $USER; cd $USER")
        if PYTHONVENV is not None:
            s.add(". {}/bin/activate".format(PYTHONVENV))
        s.add("echo")

        for xpath in file_chunk:
            fname = basename(xpath)
            s.separator(" ")
            s.separator("=")
            s.echo(f"Processing {fname}:")
            s.add("pwd")
            s.add(f"xrdcp {xpath} {fname}")
            s.add(f"ls -al {fname}")
            s.add("km3pipe --version")
            s.add(f"KPrintTree -f {fname}")
            out_fname = fname + SUFFIX
            out_fpath = join(OUTPUT_PATH, out_fname)
            tmp_fname = out_fname + ".copying"
            tmp_fpath = join(OUTPUT_PATH, tmp_fname)
            s.add(f"{SCRIPT} {fname} -o {out_fname}")
            s.cp(out_fname, tmp_fpath)
            s.add(f"rm {out_fname}")
            s.mv(tmp_fpath, out_fpath)
            s.echo(f"File '{fname}' processed.")
            s.separator("-")

        walltime = time.strftime("%H:%M:%S", time.gmtime(ET_PER_FILE * n_files))

        kp.shell.qsub(
            s,
            "{}_{}".format(JOB_NAME, job_id),
            walltime=walltime,
            fsize=FSIZE,
            vmem=VMEM,
            log_path=LOG_PATH,
            xrootd=True,
            cluster=CLUSTER,
            dryrun=DRYRUN,
        )

        if DRYRUN:
            break

        s.clear()


if __name__ == "__main__":
    main()



# Filename: clb.py
"""
Pumps for the CLB data formats.

"""

import struct
from struct import unpack

import numpy as np

from km3pipe.dataclasses import Table
from km3pipe.core import Blob, Pump
from km3pipe.sys import ignored

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2016, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"


class CLBPump(Pump):
    """A pump for binary CLB files.

    Parameters
    ----------
    file: str
        filename or file-like object.

    """

    pmt_dt = np.dtype([("channel_id", np.uint8), ("time", ">i"), ("tot", np.uint8)])

    def configure(self):
        self.file = self.require("file")
        if isinstance(self.file, str):
            self.file = open(self.file, "rb")
        self._packet_positions = []

        self._determine_packet_positions()

        self.blobs = self.blob_generator()

    def _determine_packet_positions(self):
        """Record the file pointer position of each frame"""
        self.cprint("Scanning UDP packets...")
        self.file.seek(0, 0)
        with ignored(struct.error):
            while True:
                pointer_position = self.file.tell()
                length = unpack("<i", self.file.read(4))[0]
                self._packet_positions.append(pointer_position)
                self.file.seek(length, 1)
        self.file.seek(0, 0)

    def __len__(self):
        return len(self._packet_positions)

    def seek_to_packet(self, index):
        """Move file pointer to the packet with given index."""
        pointer_position = self._packet_positions[index]
        self.file.seek(pointer_position, 0)

    def blob_generator(self):
        """Generate next blob in file"""
        for _ in range(len(self)):
            yield self.extract_blob()

    def extract_blob(self):
        try:
            length = unpack("<i", self.file.read(4))[0]
        except struct.error:
            raise StopIteration

        blob = Blob()

        blob["PacketInfo"] = Table(
            {
                "data_type": b"".join(unpack("cccc", self.file.read(4))).decode(),
                "run": unpack(">i", self.file.read(4))[0],
                "udp_sequence": unpack(">i", self.file.read(4))[0],
                "timestamp": unpack(">I", self.file.read(4))[0],
                "ns_ticks": unpack(">I", self.file.read(4))[0],
                "dom_id": unpack(">i", self.file.read(4))[0],
                "dom_status": unpack(">I", self.file.read(4))[0],
            },
            h5loc="/packet_info",
            split_h5=True,
            name="UDP Packet Info",
        )

        remaining_length = length - 7 * 4
        pmt_data = []

        count = remaining_length // self.pmt_dt.itemsize

        pmt_data = np.fromfile(self.file, dtype=self.pmt_dt, count=count)

        blob["Hits"] = Table(pmt_data, h5loc="/hits", split_h5=True)
        return blob

    def __getitem__(self, index):
        """Return blob at given index."""
        self.seek_to_packet(index)
        return self.extract_blob()

    def process(self, blob):
        return next(self.blobs)

    def __iter__(self):
        self.file.seek(0, 0)
        self.blobs = self.blob_generator()
        return self

    def __next__(self):
        return next(self.blobs)

    def finish(self):
        """Clean everything up"""
        self.file.close()



# Filename: runtable.py
"""
Prints the run table for a given detector ID.

Usage:
    runtable [options] DET_ID
    runtable (-h | --help)
    runtable --version

Options:
    -h --help           Show this screen.
    -c                  Compact view.
    -n RUNS             Number of runs.
    -r FROM_RUN-TO_RUN  Range of runs (example: 3100-3200).
    -t TARGET           Job target (run/on/off)
    -s REGEX            Regular expression to filter the runsetup name/id.
    DET_ID              Detector ID (eg. D_ARCA001).

"""

import re
import sys
import km3pipe as kp

__author__ = "Tamas Gal"
__copyright__ = "Copyright 2018, Tamas Gal and the KM3NeT collaboration."
__credits__ = []
__license__ = "MIT"
__maintainer__ = "Tamas Gal"
__email__ = "tgal@km3net.de"
__status__ = "Development"

log = kp.logger.get_logger(__name__)


def runtable(
    det_id, n=5, run_range=None, target=None, compact=False, sep="\t", regex=None
):
    """Print the run table of the last `n` runs for given detector"""
    db = kp.db.DBManager()
    df = db.run_table(det_id)

    if run_range is not None:
        try:
            from_run, to_run = [int(r) for r in run_range.split("-")]
        except ValueError:
            log.critical("Please specify a valid range (e.g. 3100-3200)!")
            raise SystemExit
        else:
            df = df[(df.RUN >= from_run) & (df.RUN <= to_run)]

    if regex is not None:
        try:
            re.compile(regex)
        except re.error:
            log.error("Invalid regex!")
            return

        df = df[
            df["RUNSETUPNAME"].str.contains(regex)
            | df["RUNSETUPID"].str.contains(regex)
        ]

    if target is not None:
        df = df[df["JOBTARGET"] == target.capitalize()]

    if n is not None:
        df = df.tail(n)

    if compact:
        df = df[["RUN", "DATETIME", "RUNSETUPNAME"]]

    df.to_csv(sys.stdout, sep=sep)


def main():
    from docopt import docopt

    args = docopt(__doc__, version=kp.version)

    try:
        n = int(args["-n"])
    except TypeError:
        n = None

    runtable(
        args["DET_ID"],
        n=n,
        run_range=args["-r"],
        target=args["-t"],
        regex=args["-s"],
        compact=args["-c"],
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# vim: ts=4 sw=4 et
"""
=======================
K40 Calibration Summary
=======================

Combine k40calib results into a single CSV file.

.. code-block:: bash

    Usage:
        k40summary.py CALIB_FOLDER
        k40summary.py (-h | --help)

    Options:
        -h --help   Show this screen.

"""

from glob import glob
import os
import re
import pickle
import numpy as np


def write_header(fobj):
    """Add the header to the CSV file"""
    fobj.write("# K40 calibration results\n")
    fobj.write("det_id\trun_id\tdom_id")
    for param in ["t0", "qe"]:
        for i in range(31):
            fobj.write("\t{}_ch{}".format(param, i))


def main():
    from docopt import docopt

    args = docopt(__doc__)

    file_pattern = os.path.join(args["CALIB_FOLDER"], "*.p")
    files = glob(file_pattern)

    with open("k40calib_summary.csv", "w") as csv_file:
        write_header(csv_file)

        for fn in files:
            det_id, run_id = [int(x) for x in re.search("_(\\d{8})" * 2, fn).groups()]
            with open(fn, "rb") as f:
                data = pickle.load(f)
            if not data:
                print("Empty dataset found for '{}'".format(fn))
            else:
                for dom_id in data.keys():
                    t0s = data[dom_id]["opt_t0s"].x
                    qes = data[dom_id]["opt_qes"].x
                    cols = np.concatenate([[det_id, run_id, dom_id], t0s, qes])
                    line = "\n" + "\t".join(str(c) for c in cols)
                    csv_file.write(line)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# vim: ts=4 sw=4 et
# Author: Tamás Gál (tgal@km3net.de)
"""
JLigier MSG log dumper.

Usage:
    msg_logger.py [-i SECONDS] [-p PATH] DET_SN

Options:
    -h --help       Show this screen.
    -i SECONDS      Write interval in seconds [Default: 5].
    -p PATH         Target path to store the log files [Default: .].
    DET_SN          Detector serial number (eg. 14).

"""

import threading
from time import sleep
import os
import urllib2
import json

import km3pipe as km3
from km3pipe.logger import get_logger

from pyslack import SlackClient

__author__ = "tamasgal"

log = get_logger(__name__)
RUN_NUMBER_URL = "http://192.168.0.120:1301/mon/controlunit/runnumber"


class MessageDumper(km3.Module):
    def configure(self):
        self.write_interval = self.get("write_interval") or 1
        self.path = self.get("path") or "."
        self.det_sn = self.get("det_sn")
        self.messages = []
        self.slack = SlackClient(km3.config.Config().slack_token)
        self.cuckoo = km3.tools.Cuckoo(interval=5, callback=self.send_message)

        if not os.path.exists(self.path):
            os.makedirs(self.path)

        self.thread = None
        self.lock = threading.Lock()
        self._start_thread()

    @property
    def target_file(self):
        current_run = int(json.load(urllib2.urlopen(RUN_NUMBER_URL))["value"])
        return os.path.join(
            self.path,
            "MSG_dump_{0:08}_{1:08}.log".format(int(self.det_sn), current_run),
        )

    def process(self, blob):
        message = blob["CHData"]
        if "ERROR" in message:
            self.cuckoo.msg(message)
            log.error(message)
        with self.lock:
            self.messages.append(message)
        return blob

    def send_message(self, message):
        self.slack.chat_post_message("#live-arca-it", message, username="ligier")

    def _start_thread(self):
        self.thread = threading.Thread(target=self._run, args=())
        self.thread.daemon = True
        self.thread.start()

    def _run(self):
        while True:
            self._dump(self.target_file)
            sleep(self.write_interval)

    def _dump(self, filename):
        with self.lock:
            print("Dumping {0} messages to {1}".format(len(self.messages), filename))
            with open(filename, "a+") as f:
                for message in self.messages:
                    f.write(message + "\n")
            self.messages = []

    def finish(self):
        self._dump(self.target_file)


def main(det_sn, target_path, write_interval):
    pipe = km3.Pipeline()
    pipe.attach(
        km3.io.CHPump,
        host="localhost",
        port=5553,
        tags="MSG",
        timeout=60 * 60 * 24 * 7,
        max_queue=2000,
    )
    pipe.attach(
        MessageDumper, write_interval=write_interval, det_sn=det_sn, path=target_path
    )
    pipe.drain()


if __name__ == "__main__":
    from docopt import docopt

    arguments = docopt(__doc__, version=1.0)

    main(arguments["DET_SN"], arguments["-p"], int(arguments["-i"]))



#!/usr/bin/env python
# vim: ts=4 sw=4 et

from datetime import datetime
from collections import deque, defaultdict
from functools import partial
from io import StringIO
from queue import Queue, Empty
import os
import shutil
import time
import threading

import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import matplotlib.dates as md
from matplotlib.colors import LogNorm
from matplotlib import pylab
import pandas as pd
import numpy as np

from km3pipe import Pipeline, Module
from km3pipe.calib import Calibration
from km3pipe.dataclasses import HitSeries
from km3pipe.io import CHPump
from km3pipe.io.daq import DAQProcessor, DAQPreamble, DAQSummaryslice, DAQEvent
from km3pipe.time import tai_timestamp
import km3pipe.style

km3pipe.style.use("km3pipe")

PLOTS_PATH = "/home/km3net/monitoring/www/plots"
N_DOMS = 18
N_DUS = 2
cal = Calibration(det_id=14)
detector = cal.detector

xfmt = md.DateFormatter("%Y-%m-%d %H:%M")
lock = threading.Lock()


class DOMHits(Module):
    def configure(self):
        self.run = True
        self.max_events = 1000
        self.hits = deque(maxlen=1000)
        self.triggered_hits = deque(maxlen=1000)
        self.thread = threading.Thread(target=self.plot).start()

    def process(self, blob):
        tag = str(blob["CHPrefix"].tag)

        if not tag == "IO_EVT":
            return blob

        data = blob["CHData"]
        data_io = StringIO(data)
        preamble = DAQPreamble(file_obj=data_io)  # noqa
        event = DAQEvent(file_obj=data_io)
        with lock:
            hits = np.zeros(N_DOMS * N_DUS)
            for dom_id, _, _, _ in event.snapshot_hits:
                du, floor, _ = detector.doms[dom_id]
                hits[(du - 1) * N_DOMS + floor - 1] += 1
            self.hits.append(hits)
            triggered_hits = np.zeros(N_DOMS * N_DUS)
            for dom_id, _, _, _, _ in event.triggered_hits:
                du, floor, _ = detector.doms[dom_id]
                triggered_hits[(du - 1) * N_DOMS + floor - 1] += 1
            self.triggered_hits.append(triggered_hits)

        return blob

    def plot(self):
        while self.run:
            with lock:
                self.create_plots()
            time.sleep(50)

    def create_plots(self):
        if len(self.hits) > 0:
            self.create_plot(self.hits, "Hits on DOMs", "hits_on_doms")
        if len(self.triggered_hits) > 0:
            self.create_plot(
                self.triggered_hits, "Triggered Hits on DOMs", "triggered_hits_on_doms"
            )

    def create_plot(self, hits, title, filename):
        fig, ax = plt.subplots(figsize=(16, 8))
        ax.grid(True)
        ax.set_axisbelow(True)
        hit_matrix = np.array([np.array(x) for x in hits]).transpose()
        im = ax.matshow(
            hit_matrix,
            interpolation="nearest",
            filternorm=None,
            cmap="plasma",
            aspect="auto",
            origin="lower",
            zorder=3,
            norm=LogNorm(vmin=1, vmax=np.amax(hit_matrix)),
        )
        yticks = np.arange(N_DOMS * N_DUS)
        ytick_labels = [
            "DU{0:0.0f}-DOM{1:02d}".format(np.ceil((y + 1) / N_DOMS), y % (N_DOMS) + 1)
            for y in yticks
        ]
        ax.set_yticks(yticks)
        ax.set_yticklabels(ytick_labels)
        ax.tick_params(labelbottom=False)
        ax.tick_params(labeltop=False)
        ax.set_xlabel("event (latest on the right)")
        ax.set_title(
            "{0} - via the last {1} Events\n{2}".format(
                title, self.max_events, datetime.utcnow().strftime("%c")
            )
        )
        cb = fig.colorbar(im, pad=0.05)
        cb.set_label("number of hits")

        fig.tight_layout()

        f = os.path.join(PLOTS_PATH, filename + ".png")
        f_tmp = os.path.join(PLOTS_PATH, filename + "_tmp.png")
        plt.savefig(f_tmp, dpi=120, bbox_inches="tight")
        plt.close("all")
        shutil.move(f_tmp, f)

    def finish(self):
        self.run = False
        if self.thread is not None:
            self.thread.stop()


class TriggerRate(Module):
    def configure(self):
        self.run = True
        self.interval = 60
        self.event_times = deque(maxlen=4000)  # max events per interval
        self.trigger_rates = deque(maxlen=60 * 48)  # minutes
        self.thread = threading.Thread(target=self.plot).start()
        self.store = pd.HDFStore("data/trigger_rates.h5", "r")
        self.restore_data()

    def restore_data(self):
        with lock:
            data = zip(
                self.store.trigger_rates.timestamp, self.store.trigger_rates.rate
            )
            self.trigger_rates.extend(data)
            print("{0} data points restored.".format(len(self.trigger_rates)))

    def process(self, blob):
        tag = str(blob["CHPrefix"].tag)

        if not tag == "IO_EVT":
            return blob

        data = blob["CHData"]
        data_io = StringIO(data)
        preamble = DAQPreamble(file_obj=data_io)  # noqa
        event = DAQEvent(file_obj=data_io)
        timestamp = event.header.time_stamp
        with lock:
            self.event_times.append(timestamp)

        return blob

    def plot(self):
        while self.run:
            time.sleep(self.interval)
            with lock:
                self.create_plot()

    def create_plot(self):
        print(self.__class__.__name__ + ": updating plot.")
        timestamp = time.time()
        now = datetime.utcnow()
        interval = self.interval
        n_events = sum(t > timestamp - interval for t in self.event_times)
        rate = n_events / interval
        self.trigger_rates.append((now, rate))
        try:
            self.store.append(
                "trigger_rates", pd.DataFrame({"timestamp": [now], "rate": [rate]})
            )
        except pd.io.pytables.ClosedFileError:
            pass
        print(
            "Number of rates recorded: {0} (last: {1}".format(
                len(self.trigger_rates), self.trigger_rates[-1]
            )
        )

        x, y = zip(*self.trigger_rates)
        if not any(y):
            print("Waiting for more values...")
            return
        fig, ax = plt.subplots(figsize=(16, 4))
        ax.xaxis.set_major_formatter(xfmt)
        data = pd.DataFrame({"dates": x, "rates": y})
        #            plt.scatter(x, y)
        data.plot("dates", "rates", grid=True, ax=ax, legend=False, style=".")
        ax.set_title(
            "Trigger Rate - via Event Times\n{0}".format(
                datetime.utcnow().strftime("%c")
            )
        )
        ax.set_xlabel("time")
        ax.set_ylabel("trigger rate [Hz]")
        #        ax.set_ylim(-0.1)
        try:
            ax.set_yscale("log")
        except ValueError:
            pass

        fig.tight_layout()

        filename = os.path.join(PLOTS_PATH, "trigger_rates.png")
        filename_tmp = os.path.join(PLOTS_PATH, "trigger_rates_tmp.png")
        plt.savefig(filename_tmp, dpi=120, bbox_inches="tight")
        plt.close("all")
        shutil.move(filename_tmp, filename)

    def finish(self):
        self.run = False
        if self.thread is not None:
            self.thread.stop()
        if self.store.is_open:
            self.store.close()


class DOMActivityPlotter(Module):
    def configure(self):
        self.index = 0
        self.rates = defaultdict(partial(deque, maxlen=4000))
        self.run = True
        self.thread = threading.Thread(target=self.plot, args=()).start()

    def process(self, blob):
        self.index += 1
        if self.index % 30:
            return blob

        tag = str(blob["CHPrefix"].tag)
        data = blob["CHData"]

        if not tag == "IO_SUM":
            return blob

        data = blob["CHData"]
        data_io = StringIO(data)
        preamble = DAQPreamble(file_obj=data_io)  # noqa
        summaryslice = DAQSummaryslice(file_obj=data_io)
        timestamp = summaryslice.header.time_stamp
        with lock:
            for dom_id, rates in summaryslice.summary_frames.iteritems():
                du, dom, _ = detector.doms[dom_id]
                self.rates[(du, dom)].append((timestamp, sum(rates)))

        return blob

    def plot(self):
        while self.run:
            with lock:
                self.create_plot()
            time.sleep(50)

    def create_plot(self):
        print(self.__class__.__name__ + ": updating plot.")
        x, y, _ = zip(*detector.doms.values())
        fig, ax = plt.subplots(figsize=(10, 6))
        cmap = plt.get_cmap("RdYlGn_r")
        cmap.set_over("deeppink", 1.0)
        cmap.set_under("deepskyblue", 1.0)

        vmax = 15 * 60

        scatter_args = {
            "edgecolors": "None",
            "s": 100,
            "vmin": 0.0,
            "vmax": vmax,
        }
        sc_inactive = ax.scatter(x, y, c="lightgray", label="inactive", **scatter_args)
        now = tai_timestamp()

        try:
            xa, ya = map(np.array, zip(*self.rates.keys()))
            ts = np.array([now - max(zip(*d)[0]) for d in self.rates.values()])
        except ValueError:
            print("Not enough data.")
            pass
        else:
            # print(self.__class__.__name__ + ": updating plot.")
            active_idx = ts < vmax
            sc_active = ax.scatter(
                xa[active_idx],
                ya[active_idx],
                c=ts[active_idx],
                cmap=cmap,
                **scatter_args
            )
            ax.scatter(
                xa[~active_idx],
                ya[~active_idx],
                c="deeppink",
                label="> {0} s".format(vmax),
                **scatter_args
            )
            cb = plt.colorbar(sc_active)
            cb.set_label("last activity [s]")
            # ts_series = pd.Series(ts)
            # print(ts_series.describe())

        ax.set_title(
            "DOM Activity - via Summary Slices\n{0}".format(
                datetime.utcnow().strftime("%c")
            )
        )
        ax.set_xlabel("DU")
        ax.set_ylabel("DOM")
        ax.set_ylim(-2)
        ax.set_yticks(range(1, N_DOMS + 1))
        major_locator = pylab.MaxNLocator(integer=True)
        sc_inactive.axes.xaxis.set_major_locator(major_locator)

        ax.legend(
            bbox_to_anchor=(0.0, -0.16, 1.0, 0.102),
            loc=1,
            ncol=2,
            mode="expand",
            borderaxespad=0.0,
        )

        fig.tight_layout()

        filename = os.path.join(PLOTS_PATH, "dom_activity.png")
        filename_tmp = os.path.join(PLOTS_PATH, "dom_activity_tmp.png")
        plt.savefig(filename_tmp, dpi=120, bbox_inches="tight")
        plt.close("all")
        shutil.move(filename_tmp, filename)

    def finish(self):
        self.run = False
        if self.thread is not None:
            self.thread.stop()


class ZTPlot(Module):
    def configure(self):
        self.run = True
        self.max_queue = 3
        self.queue = Queue()
        self.thread = threading.Thread(target=self.plot).start()

    def process(self, blob):
        if "Hits" not in blob:
            return blob

        hits = blob["Hits"]
        e_info = blob["EventInfo"]

        print("Event queue size: {0}".format(self.queue.qsize()))
        if self.queue.qsize() < self.max_queue:
            self.queue.put((e_info, hits))

        return blob

    def plot(self):
        while self.run:
            try:
                e_info, hits = self.queue.get(timeout=50)
            except Empty:
                continue
            with lock:
                self.create_plot(e_info, hits)

    def create_plot(self, e_info, hits):
        print(self.__class__.__name__ + ": updating plot.")
        cal.apply(hits)
        dus = set(detector.doms[h.dom_id][0] for h in hits)

        n_plots = len(dus)
        n_cols = int(np.ceil(np.sqrt(n_plots)))
        n_rows = int(n_plots / n_cols) + (n_plots % n_cols > 0)
        fig, axes = plt.subplots(
            ncols=n_cols, nrows=n_rows, sharex=True, sharey=True, figsize=(16, 8)
        )

        for ax, du in zip(axes.flat, dus):
            _hits = [h for h in hits if detector.doms[h.dom_id][0] == du]
            du_hits = HitSeries(_hits)
            trig_hits = HitSeries([h for h in _hits if h.triggered])

            ax.scatter(
                du_hits.time, [z for (x, y, z) in du_hits.pos], c="#09A9DE", label="hit"
            )
            ax.scatter(
                trig_hits.time,
                [z for (x, y, z) in trig_hits.pos],
                c="#FF6363",
                label="triggered hit",
            )
            ax.set_title("DU{0}".format(du), fontsize=16, fontweight="bold")

        for ax in axes.flat:
            ax.tick_params(labelsize=16)
            ax.yaxis.set_major_locator(ticker.MultipleLocator(200))
            xlabels = ax.get_xticklabels()
            for label in xlabels:
                label.set_rotation(45)

        plt.suptitle(
            "Run {0}, FrameIndex {1}, TriggerCounter {2}\n{3}".format(
                e_info.run_id,
                e_info.frame_index,
                e_info.trigger_counter,
                datetime.utcfromtimestamp(e_info.utc_seconds),
            ),
            fontsize=16,
        )
        fig.text(0.5, 0.01, "time [ns]", ha="center")
        fig.text(0.08, 0.5, "z [m]", va="center", rotation="vertical")
        #        plt.tight_layout()

        filename = "ztplot"
        f = os.path.join(PLOTS_PATH, filename + ".png")
        f_tmp = os.path.join(PLOTS_PATH, filename + "_tmp.png")
        plt.savefig(f_tmp, dpi=120, bbox_inches="tight")
        plt.close("all")
        shutil.move(f_tmp, f)

    def finish(self):
        self.run = False
        if self.thread is not None:
            self.thread.stop()


pipe = Pipeline()
pipe.attach(
    CHPump,
    host="192.168.0.110",
    port=5553,
    tags="IO_EVT, IO_SUM",
    timeout=60 * 60 * 24 * 7,
    max_queue=2000,
)
pipe.attach(DAQProcessor)
pipe.attach(DOMActivityPlotter)
pipe.attach(TriggerRate)
pipe.attach(DOMHits)
pipe.attach(ZTPlot)
pipe.drain()
